class _Application:
    def __init__(self):
        self._adaptable = None
        self._amazon_aws = None
        self._apache = None
        self._azure_key_vault = None
        self._basic = None
        self._blue_coat_sslva = None
        self._capi = None
        self._citrix_net_scaler = None
        self._connect_direct = None
        self._f5_authentication_bundle = None
        self._f5_ltm_advanced = None
        self._google_cloud_load_balancer = None
        self._ibm_datapower = None
        self._ibm_gsk = None
        self._imperva_mx = None
        self._jks = None
        self._juniper_sas = None
        self._oracle_iplanet = None
        self._palo_alto_network_fw = None
        self._pem = None
        self._pkcs11 = None
        self._pkcs12 = None
        self._riverbed_steel_head = None
        self._tealeaf_pca = None
        self._vamnshield = None

    @property
    def adaptable(self):
        if not self._adaptable:
            from pytpp.attributes.adaptable_app import AdaptableAppAttributes
            self._adaptable = AdaptableAppAttributes
        return self._adaptable

    @property
    def amazon_aws(self):
        if not self._amazon_aws:
            from pytpp.attributes.amazon_app import AmazonAppAttributes
            self._amazon_aws = AmazonAppAttributes
        return self._amazon_aws

    @property
    def apache(self):
        if not self._apache:
            from pytpp.attributes.apache import ApacheAttributes
            self._apache = ApacheAttributes
        return self._apache

    @property
    def azure_key_vault(self):
        if not self._azure_key_vault:
            from pytpp.attributes.azure_key_vault import AzureKeyVaultAttributes
            self._azure_key_vault = AzureKeyVaultAttributes
        return self._azure_key_vault

    @property
    def basic(self):
        if not self._basic:
            from pytpp.attributes.basic import BasicAttributes
            self._basic = BasicAttributes
        return self._basic

    @property
    def blue_coat_sslva(self):
        if not self._blue_coat_sslva:
            from pytpp.attributes.bluecoat_sslva import BlueCoatSSLVAAttributes
            self._blue_coat_sslva = BlueCoatSSLVAAttributes
        return self._blue_coat_sslva

    @property
    def capi(self):
        if not self._capi:
            from pytpp.attributes.capi import CAPIAttributes
            self._capi = CAPIAttributes
        return self._capi

    @property
    def citrix_net_scaler(self):
        if not self._citrix_net_scaler:
            from pytpp.attributes.netscaler import NetScalerAttributes
            self._citrix_net_scaler = NetScalerAttributes
        return self._citrix_net_scaler

    @property
    def connect_direct(self):
        if not self._connect_direct:
            from pytpp.attributes.connectdirect import ConnectDirectAttributes
            self._connect_direct = ConnectDirectAttributes
        return self._connect_direct

    @property
    def f5_authentication_bundle(self):
        if not self._f5_authentication_bundle:
            from pytpp.attributes.f5_authentication_bundle import F5AuthenticationBundleAttributes
            self._f5_authentication_bundle = F5AuthenticationBundleAttributes
        return self._f5_authentication_bundle

    @property
    def f5_ltm_advanced(self):
        if not self._f5_ltm_advanced:
            from pytpp.attributes.f5_ltm_advanced import F5LTMAdvancedAttributes
            self._f5_ltm_advanced = F5LTMAdvancedAttributes
        return self._f5_ltm_advanced

    @property
    def google_cloud_load_balancer(self):
        if not self._google_cloud_load_balancer:
            from pytpp.attributes.google_cloud_app import GoogleCloudAppAttributes
            self._google_cloud_load_balancer = GoogleCloudAppAttributes
        return self._google_cloud_load_balancer

    @property
    def ibm_datapower(self):
        if not self._ibm_datapower:
            from pytpp.attributes.datapower import DataPowerAttributes
            self._ibm_datapower = DataPowerAttributes
        return self._ibm_datapower

    @property
    def ibm_gsk(self) :
        if not self._ibm_gsk:
            from pytpp.attributes.gsk import GSKAttributes
            self._ibm_gsk = GSKAttributes
        return self._ibm_gsk

    @property
    def imperva_mx(self) :
        if not self._imperva_mx:
            from pytpp.attributes.imperva_mx import ImpervaMXAttributes
            self._imperva_mx = ImpervaMXAttributes
        return self._imperva_mx

    @property
    def jks(self) :
        if not self._jks:
            from pytpp.attributes.jks import JKSAttributes
            self._jks = JKSAttributes
        return self._jks

    @property
    def juniper_sas(self) :
        if not self._juniper_sas:
            from pytpp.attributes.juniper_sas import JuniperSASAttributes
            self._juniper_sas = JuniperSASAttributes
        return self._juniper_sas

    @property
    def oracle_iplanet(self) :
        if not self._oracle_iplanet:
            from pytpp.attributes.iplanet import iPlanetAttributes
            self._oracle_iplanet = iPlanetAttributes
        return self._oracle_iplanet

    @property
    def palo_alto_network_fw(self) :
        if not self._palo_alto_network_fw:
            from pytpp.attributes.palo_alto_network_fw import PaloAltoNetworkFWAttributes
            self._palo_alto_network_fw = PaloAltoNetworkFWAttributes
        return self._palo_alto_network_fw

    @property
    def pem(self) :
        if not self._pem:
            from pytpp.attributes.pem import PEMAttributes
            self._pem = PEMAttributes
        return self._pem

    @property
    def pkcs11(self) :
        if not self._pkcs11:
            from pytpp.attributes.pkcs11 import PKCS11Attributes
            self._pkcs11 = PKCS11Attributes
        return self._pkcs11

    @property
    def pkcs12(self) :
        if not self._pkcs12:
            from pytpp.attributes.pkcs_12 import PKCS12Attributes
            self._pkcs12 = PKCS12Attributes
        return self._pkcs12

    @property
    def riverbed_steel_head(self) :
        if not self._riverbed_steel_head:
            from pytpp.attributes.riverbed_steelhead import RiverbedSteelHeadAttributes
            self._riverbed_steel_head = RiverbedSteelHeadAttributes
        return self._riverbed_steel_head

    @property
    def tealeaf_pca(self) :
        if not self._tealeaf_pca:
            from pytpp.attributes.tealeaf_pca import TealeafPCAAttributes
            self._tealeaf_pca = TealeafPCAAttributes
        return self._tealeaf_pca

    @property
    def vamnshield(self) :
        if not self._vamnshield:
            from pytpp.attributes.vam_nshield import VAMnShieldAttributes
            self._vamnshield = VAMnShieldAttributes
        return self._vamnshield


class _ApplicationGroup:
    def __init__(self):
        self._apache = None
        self._pkcs11 = None

    @property
    def apache(self) :
        if not self._apache:
            from pytpp.attributes.apache_application_group import ApacheApplicationGroupAttributes
            self._apache = ApacheApplicationGroupAttributes
        return self._apache

    @property
    def pkcs11(self) :
        if not self._pkcs11:
            from pytpp.attributes.pkcs11_application_group import PKCS11ApplicationGroupAttributes
            self._pkcs11 = PKCS11ApplicationGroupAttributes
        return self._pkcs11


class _CertificateAuthority:
    def __init__(self):
        self._adaptable = None
        self._msca = None
        self._self_signed = None

    @property
    def adaptable(self) :
        if not self._adaptable:
            from pytpp.attributes.adaptable_ca import AdaptableCAAttributes
            self._adaptable = AdaptableCAAttributes
        return self._adaptable

    @property
    def msca(self) :
        if not self._msca:
            from pytpp.attributes.microsoft_ca import MicrosoftCAAttributes
            self._msca = MicrosoftCAAttributes
        return self._msca

    @property
    def self_signed(self) :
        if not self._self_signed:
            from pytpp.attributes.self_signed_ca import SelfSignedCAAttributes
            self._self_signed = SelfSignedCAAttributes
        return self._self_signed


class _ClientWork:
    def __init__(self):
        self._agent_connectivity = None
        self._agent_upgrade = None
        self._certificate_device_placement = None
        self._certificate_discovery = None
        self._certificate_enrollment_via_est_protocol = None
        self._certificate_installation = None
        self._device_certificate_creation = None
        self._dynamic_provisioning = None
        self._ssh_device_placement = None
        self._ssh_discovery = None
        self._ssh_key_usage = None
        self._ssh_remediation = None
        self._user_certificate_creation = None

    @property
    def agent_connectivity(self) :
        if not self._agent_connectivity:
            from pytpp.attributes.client_agent_configuration_work import ClientAgentConfigurationWorkAttributes
            self._agent_connectivity = ClientAgentConfigurationWorkAttributes
        return self._agent_connectivity

    @property
    def agent_upgrade(self):
        if not self._agent_upgrade:
            from pytpp.attributes.client_agent_automatic_upgrade_work import ClientAgentAutomaticUpgradeWorkAttributes
            self._agent_upgrade = ClientAgentAutomaticUpgradeWorkAttributes
        return self._agent_upgrade

    @property
    def certificate_device_placement(self) :
        if not self._certificate_device_placement:
            from pytpp.attributes.server_agent_cert_device_placement_work import ServerAgentCertDevicePlacementWorkAttributes
            self._certificate_device_placement = ServerAgentCertDevicePlacementWorkAttributes
        return self._certificate_device_placement

    @property
    def certificate_discovery(self) :
        if not self._certificate_discovery:
            from pytpp.attributes.client_certificate_discovery_work import ClientCertificateDiscoveryWorkAttributes
            self._certificate_discovery = ClientCertificateDiscoveryWorkAttributes
        return self._certificate_discovery

    @property
    def certificate_enrollment_via_est_protocol(self) :
        if not self._certificate_enrollment_via_est_protocol:
            from pytpp.attributes.network_device_certificate_work import NetworkDeviceCertificateWorkAttributes
            self._certificate_enrollment_via_est_protocol = NetworkDeviceCertificateWorkAttributes
        return self._certificate_enrollment_via_est_protocol

    @property
    def certificate_installation(self) :
        if not self._certificate_installation:
            from pytpp.attributes.certificate_provisioning_work import CertificateProvisioningWorkAttributes
            self._certificate_installation = CertificateProvisioningWorkAttributes
        return self._certificate_installation

    @property
    def dynamic_provisioning(self) :
        if not self._dynamic_provisioning:
            from pytpp.attributes.server_certificate_work import ServerCertificateWorkAttributes
            self._dynamic_provisioning = ServerCertificateWorkAttributes
        return self._dynamic_provisioning

    @property
    def ssh_device_placement(self) :
        if not self._ssh_device_placement:
            from pytpp.attributes.server_agent_ssh_device_placement_work import ServerAgentSSHDevicePlacementWorkAttributes
            self._ssh_device_placement = ServerAgentSSHDevicePlacementWorkAttributes
        return self._ssh_device_placement

    @property
    def ssh_discovery(self) :
        if not self._ssh_discovery:
            from pytpp.attributes.client_agent_ssh_discovery_work import ClientAgentSSHDiscoveryWorkAttributes
            self._ssh_discovery = ClientAgentSSHDiscoveryWorkAttributes
        return self._ssh_discovery

    @property
    def ssh_key_usage(self) :
        if not self._ssh_key_usage:
            from pytpp.attributes.client_agent_ssh_key_usage_work import ClientAgentSSHKeyUsageWorkAttributes
            self._ssh_key_usage = ClientAgentSSHKeyUsageWorkAttributes
        return self._ssh_key_usage

    @property
    def ssh_remediation(self) :
        if not self._ssh_remediation:
            from pytpp.attributes.client_agent_ssh_provisioning_work import ClientAgentSSHProvisioningWorkAttributes
            self._ssh_remediation = ClientAgentSSHProvisioningWorkAttributes
        return self._ssh_remediation

    @property
    def user_certificate_creation(self) :
        if not self._user_certificate_creation:
            from pytpp.attributes.client_user_certificate_work import ClientUserCertificateWorkAttributes
            self._user_certificate_creation = ClientUserCertificateWorkAttributes
        return self._user_certificate_creation


class _Credential:
    def __init__(self):
        self._amazon = None
        self._certificate = None
        self._generic = None
        self._password = None
        self._private_key = None
        self._upcred = None

    @property
    def amazon(self) :
        if not self._amazon:
            from pytpp.attributes.amazon_credential import AmazonCredentialAttributes
            self._amazon = AmazonCredentialAttributes
        return self._amazon

    @property
    def certificate(self) :
        if not self._certificate:
            from pytpp.attributes.certificate_credential import CertificateCredentialAttributes
            self._certificate = CertificateCredentialAttributes
        return self._certificate

    @property
    def generic(self) :
        if not self._generic:
            from pytpp.attributes.generic_credential import GenericCredentialAttributes
            self._generic = GenericCredentialAttributes
        return self._generic

    @property
    def password(self) :
        if not self._password:
            from pytpp.attributes.password_credential import PasswordCredentialAttributes
            self._password = PasswordCredentialAttributes
        return self._password

    @property
    def private_key(self) :
        if not self._private_key:
            from pytpp.attributes.private_key_credential import PrivateKeyCredentialAttributes
            self._private_key = PrivateKeyCredentialAttributes
        return self._private_key

    @property
    def username_password(self) :
        if not self._upcred:
            from pytpp.attributes.username_password_credential import UsernamePasswordCredentialAttributes
            self._upcred = UsernamePasswordCredentialAttributes
        return self._upcred


class _Discovery:
    def __init__(self):
        self._network = None

    @property
    def network(self) :
        if not self._network:
            from pytpp.attributes.discovery import DiscoveryAttributes
            self._network = DiscoveryAttributes
        return self._network


class _Identity:
    def __init__(self):
        self._group = None
        self._user = None

    @property
    def group(self) :
        if not self._group:
            from pytpp.attributes.group import GroupAttributes
            self._group = GroupAttributes
        return self._group

    @property
    def user(self) :
        if not self._user:
            from pytpp.attributes.user import UserAttributes
            self._user = UserAttributes
        return self._user


class _Platforms:
    def __init__(self):
        self._auto_layout_manager = None
        self._bulk_provisioning_manager = None
        self._ca_import_manager = None
        self._certificate_manager = None
        self._certificate_pre_enrollment = None
        self._certificate_revocation = None
        self._cloud_instance_monitor = None
        self._discovery_manager = None
        self._engines = None
        self._monitor = None
        self._onboard_discovery_manager = None
        self._reporting = None
        self._ssh_manager = None
        self._trustnet_manager = None
        self._validation_manager = None

    @property
    def auto_layout_manager(self) :
        if not self._auto_layout_manager:
            from pytpp.attributes.auto_layout_manager import AutoLayoutManagerAttributes
            self._auto_layout_manager = AutoLayoutManagerAttributes
        return self._auto_layout_manager

    @property
    def bulk_provisioning_manager(self) :
        if not self._bulk_provisioning_manager:
            from pytpp.attributes.bulk_provisioning_manager import BulkProvisioningManagerAttributes
            self._bulk_provisioning_manager = BulkProvisioningManagerAttributes
        return self._bulk_provisioning_manager

    @property
    def ca_import_manager(self) :
        if not self._ca_import_manager:
            from pytpp.attributes.ca_import_manager import CAImportManagerAttributes
            self._ca_import_manager = CAImportManagerAttributes
        return self._ca_import_manager

    @property
    def certificate_manager(self) :
        if not self._certificate_manager:
            from pytpp.attributes.certificate_manager import CertificateManagerAttributes
            self._certificate_manager = CertificateManagerAttributes
        return self._certificate_manager

    @property
    def certificate_pre_enrollment(self) :
        if not self._certificate_pre_enrollment:
            from pytpp.attributes.certificate_pre_enrollment import CertificatePreEnrollmentAttributes
            self._certificate_pre_enrollment = CertificatePreEnrollmentAttributes
        return self._certificate_pre_enrollment

    @property
    def certificate_revocation(self) :
        if not self._certificate_revocation:
            from pytpp.attributes.certificate_revocation import CertificateRevocationAttributes
            self._certificate_revocation = CertificateRevocationAttributes
        return self._certificate_revocation

    @property
    def cloud_instance_monitor(self) :
        if not self._cloud_instance_monitor:
            from pytpp.attributes.cloud_instance_monitor import CloudInstanceMonitorAttributes
            self._cloud_instance_monitor = CloudInstanceMonitorAttributes
        return self._cloud_instance_monitor

    @property
    def discovery_manager(self) :
        if not self._discovery_manager:
            from pytpp.attributes.discovery_manager import DiscoveryManagerAttributes
            self._discovery_manager = DiscoveryManagerAttributes
        return self._discovery_manager

    @property
    def engines(self):
        if not self._engines:
            from pytpp.attributes.venafi_platform import VenafiPlatformAttributes
            self._engines = VenafiPlatformAttributes
        return self._engines

    @property
    def monitor(self) :
        if not self._monitor:
            from pytpp.attributes.monitoring_module import MonitoringModuleAttributes
            self._monitor = MonitoringModuleAttributes
        return self._monitor

    @property
    def onboard_discovery_manager(self) :
        if not self._onboard_discovery_manager:
            from pytpp.attributes.onboard_discovery_manager import OnboardDiscoveryManagerAttributes
            self._onboard_discovery_manager = OnboardDiscoveryManagerAttributes
        return self._onboard_discovery_manager

    @property
    def reporting(self) :
        if not self._reporting:
            from pytpp.attributes.reporter_service_module import ReporterServiceModuleAttributes
            self._reporting = ReporterServiceModuleAttributes
        return self._reporting

    @property
    def ssh_manager(self) :
        if not self._ssh_manager:
            from pytpp.attributes.ssh_manager import SSHManagerAttributes
            self._ssh_manager = SSHManagerAttributes
        return self._ssh_manager

    @property
    def validation_manager(self) :
        if not self._validation_manager:
            from pytpp.attributes.validation_manager import ValidationManagerAttributes
            self._validation_manager = ValidationManagerAttributes
        return self._validation_manager


class _Workflow:
    def __init__(self):
        self._adaptable = None
        self._standard = None
        self._ticket = None

    @property
    def adaptable(self) :
        if not self._adaptable:
            from pytpp.attributes.adaptable_workflow import AdaptableWorkflowAttributes
            self._adaptable = AdaptableWorkflowAttributes
        return self._adaptable

    @property
    def standard(self) :
        if not self._standard:
            from pytpp.attributes.workflow import WorkflowAttributes
            self._standard = WorkflowAttributes
        return self._standard

    @property
    def ticket(self) :
        if not self._ticket:
            from pytpp.attributes.workflow_ticket import WorkflowTicketAttributes
            self._ticket = WorkflowTicketAttributes
        return self._ticket


class _Attributes:
    def __init__(self):
        self._applications = None
        self._application_group = None
        self._ca = None
        self._certificate = None
        self._client_groups = None
        self._client_work = None
        self._credentials = None
        self._custom_fields = None
        self._device = None
        self._discovery = None
        self._folder = None
        self._identity = None
        self._permissions = None
        self._placement_rules = None
        self._platforms = None
        self._workflow = None

    @property
    def application(self) -> _Application:
        self._applications = self._applications or _Application()
        return self._applications

    @property
    def application_group(self) -> _ApplicationGroup:
        self._application_group = self._application_group or _ApplicationGroup()
        return self._application_group

    @property
    def certificate(self):
        if not self._certificate:
            from pytpp.attributes.x509_certificate import X509CertificateAttributes
            self._certificate = X509CertificateAttributes
        return self._certificate

    @property
    def certificate_authority(self) -> _CertificateAuthority:
        self._ca = self._ca or _CertificateAuthority()
        return self._ca

    @property
    def client_groups(self):
        if not self._client_groups:
            from pytpp.attributes.client_group import ClientGroupAttributes
            self._client_groups = ClientGroupAttributes
        return self._client_groups

    @property
    def client_work(self) -> _ClientWork:
        self._client_work = self._client_work or _ClientWork()
        return self._client_work

    @property
    def credential(self) -> _Credential:
        self._credentials = self._credentials or _Credential()
        return self._credentials

    @property
    def custom_fields(self):
        if not self._custom_fields:
            from pytpp.attributes.metadata_list import MetadataListAttributes
            from pytpp.attributes.metadata_text import MetadataTextAttributes
            from pytpp.attributes.metadata_category import MetadataCategoryAttributes
            from pytpp.attributes.metadata_choice import MetadataChoiceAttributes
            from pytpp.attributes.metadata_datetime import MetadataDateTimeAttributes
            from pytpp.attributes.metadata_identity import MetadataIdentityAttributes
            from pytpp.attributes.metadata_base import IterableMeta

            class CustomFieldAttributes(metaclass=IterableMeta):
                list_type = MetadataListAttributes
                text_type = MetadataTextAttributes
                category_type = MetadataCategoryAttributes
                choice_type = MetadataChoiceAttributes
                datetime_type = MetadataDateTimeAttributes
                identity_type = MetadataIdentityAttributes

            self._custom_fields = CustomFieldAttributes
        return self._custom_fields

    @property
    def device(self):
        if not self._device:
            from pytpp.attributes.device import DeviceAttributes
            self._device = DeviceAttributes
        return self._device

    @property
    def discovery(self) -> _Discovery:
        self._discovery = self._discovery or _Discovery()
        return self._discovery

    @property
    def folder(self):
        if not self._folder:
            from pytpp.attributes.policy import PolicyAttributes
            self._folder = PolicyAttributes
        return self._folder

    @property
    def identity(self) -> _Identity:
        self._identity = self._identity or _Identity()
        return self._identity

    @property
    def placement_rules(self):
        if not self._placement_rules:
            from pytpp.attributes.layout_rule_base import LayoutRuleBaseAttributes
            self._placement_rules = LayoutRuleBaseAttributes
        return self._placement_rules

    @property
    def platforms(self) -> _Platforms:
        self._platforms = self._platforms or _Platforms()
        return self._platforms

    @property
    def workflow(self) -> _Workflow:
        self._workflow = self._workflow or _Workflow()
        return self._workflow


Attributes = _Attributes()



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class ApacheAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Apache"
	application_id = Attribute('Application ID', min_version='17.2')
	certificate_chain_file = Attribute('Certificate Chain File')
	certificate_file = Attribute('Certificate File')
	client_tools_path = Attribute('Client Tools Path', min_version='17.2')
	file_validation_disabled = Attribute('File Validation Disabled')
	network_validation_disabled = Attribute('Network Validation Disabled')
	ocs_identifier = Attribute('OCS Identifier', min_version='18.3')
	overwrite_existing_chain = Attribute('Overwrite Existing Chain')
	pbes2_algorithm = Attribute('PBES2 Algorithm', min_version='21.3')
	partition_password_credential = Attribute('Partition Password Credential', min_version='17.2')
	private_key_file = Attribute('Private Key File')
	private_key_label = Attribute('Private Key Label', min_version='19.1')
	private_key_location = Attribute('Private Key Location', min_version='17.2')
	protection_type = Attribute('Protection Type', min_version='18.1')
	slot_number = Attribute('Slot Number', min_version='17.2')
	softcard_identifier = Attribute('Softcard Identifier', min_version='18.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class CodeSigningRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Root"
	default_ca_container = Attribute('Default CA Container', min_version='19.2')
	default_certificate_container = Attribute('Default Certificate Container', min_version='19.2')
	default_credential_container = Attribute('Default Credential Container', min_version='19.2')
	enforce_group_roles = Attribute('Enforce Group Roles', min_version='19.2')
	environment_template_update_flow_dn = Attribute('Environment Template Update Flow DN', min_version='21.4')
	flow_instance_macro = Attribute('Flow Instance Macro', min_version='19.3')
	key_storage_location = Attribute('Key Storage Location', min_version='19.2')
	key_use_timeout = Attribute('Key Use Timeout', min_version='19.2')
	prevent_self_dealing = Attribute('Prevent Self Dealing', min_version='19.2')
	project_delete_flow_dn = Attribute('Project Delete Flow DN', min_version='21.1')
	project_description_tooltip = Attribute('Project Description Tooltip', min_version='19.2')
	request_in_progress_message = Attribute('Request In Progress Message', min_version='19.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.credential_driver_base import CredentialDriverBaseAttributes


class AdaptableCredentialConnectorAttributes(CredentialDriverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Adaptable Credential Connector"
	credential = Attribute('Credential', min_version='21.1')
	powershell_script = Attribute('PowerShell Script', min_version='21.1')
	powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id', min_version='21.1')
	web_service_url = Attribute('Web Service URL', min_version='21.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.encryption_driver import EncryptionDriverAttributes


class DPAPIEncryptionDriverAttributes(EncryptionDriverAttributes, metaclass=IterableMeta):
	__config_class__ = "DPAPI Encryption Driver"
	generation_only = Attribute('Generation Only', min_version='18.1')
	key_validation = Attribute('Key Validation')
	verigram = Attribute('VeriGram', min_version='18.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes


class MicrosoftCAAttributes(CertificateAuthorityBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Microsoft CA"
	enrollment_agent_certificate = Attribute('Enrollment Agent Certificate')
	given_name = Attribute('Given Name')
	include_cn_as_san = Attribute('Include CN as SAN')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.service_module import ServiceModuleAttributes


class CertificateManagerAttributes(ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "Certificate Manager"
	cdp_aia_verification_disabled = Attribute('CDP AIA Verification Disabled')
	certificate_api_todo_timeout = Attribute('Certificate API ToDo Timeout', min_version='21.4')
	escalation_notice_interval = Attribute('Escalation Notice Interval')
	escalation_notice_start = Attribute('Escalation Notice Start')
	expiration_notice_interval = Attribute('Expiration Notice Interval')
	expiration_notice_start = Attribute('Expiration Notice Start')
	maximum_threads = Attribute('Maximum Threads')
	minimum_threads = Attribute('Minimum Threads')
	renewal_window = Attribute('Renewal Window')
	renewal_window_event = Attribute('Renewal Window Event')
	revocation_check_disabled = Attribute('Revocation Check Disabled')
	service_module_classes = Attribute('Service Module Classes')
	start_time = Attribute('Start Time')
	trust_store_management_disabled = Attribute('Trust Store Management Disabled')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class ClientRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Root"
	client_enrollment_anonymous_allowed = Attribute('Client Enrollment Anonymous Allowed')
	client_enrollment_default_owner = Attribute('Client Enrollment Default Owner')
	client_enrollment_secret_allowed = Attribute('Client Enrollment Secret Allowed')
	client_enrollment_secret_dn = Attribute('Client Enrollment Secret DN')
	client_portal_signing_certificate = Attribute('Client Portal Signing Certificate', min_version='15.2')
	client_trust_certificate_dn = Attribute('Client Trust Certificate DN')
	client_trust_update = Attribute('Client Trust Update')
	dsn = Attribute('DSN')
	driver_name = Attribute('Driver Name')
	environment_variables = Attribute('Environment Variables')
	identity_variables = Attribute('Identity Variables')
	mac_ignore_list = Attribute('MAC Ignore List', min_version='15.2')
	options = Attribute('Options')
	rolling_code_tolerance = Attribute('Rolling Code Tolerance', min_version='15.4')
	rule = Attribute('Rule')
	trust_level = Attribute('Trust Level')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class BrocadeAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Brocade"
	all_cipher_suites = Attribute('All Cipher Suites')
	certificate_file = Attribute('Certificate File')
	chain_cert = Attribute('Chain Cert')
	file_validation_disabled = Attribute('File Validation Disabled')
	network_validation_disabled = Attribute('Network Validation Disabled')
	private_key_file = Attribute('Private Key File')
	privileged_mode_credential = Attribute('Privileged Mode Credential')
	ssl_profile_name = Attribute('SSL Profile Name')
	selected_cipher_suites = Attribute('Selected Cipher Suites')
	session_cache = Attribute('Session Cache')
	temp_certificate_label = Attribute('Temp Certificate Label')
	virtual_server_name = Attribute('Virtual Server Name')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.legacy_key_base import LegacyKeyBaseAttributes
from pytpp.attributes.monitoring_base import MonitoringBaseAttributes
from pytpp.attributes.top import TopAttributes


class SymmetricKeyAttributes(LegacyKeyBaseAttributes, MonitoringBaseAttributes, TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Symmetric Key"
	check_value = Attribute('Check Value')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes


class MicrosoftCAPoolAttributes(CertificateAuthorityBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Microsoft CA Pool"
	certificate_authority = Attribute('Certificate Authority', min_version='19.1')
	second_certificate_authority = Attribute('Second Certificate Authority', min_version='19.1')
	third_certificate_authority = Attribute('Third Certificate Authority', min_version='19.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.log_channel import LogChannelAttributes


class LogSNMPAttributes(LogChannelAttributes, metaclass=IterableMeta):
	__config_class__ = "Log SNMP"
	community = Attribute('Community')
	message_body = Attribute('Message Body')
	snmp_authentication_password = Attribute('SNMP Authentication Password')
	snmp_authentication_protocol = Attribute('SNMP Authentication Protocol')
	snmp_engine_id = Attribute('SNMP Engine ID')
	snmp_privacy_password = Attribute('SNMP Privacy Password')
	snmp_privacy_protocol = Attribute('SNMP Privacy Protocol')
	snmp_username = Attribute('SNMP Username')
	snmp_v3_mode = Attribute('SNMP V3 Mode')
	snmp_version = Attribute('SNMP Version')
	target = Attribute('Target')
	trap_oid = Attribute('Trap OID')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.flow_action_base import FlowActionBaseAttributes


class CodeSigningEnvironmentTemplateUpdateActionAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Environment Template Update Action"



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.service_module import ServiceModuleAttributes


class ClientGroupMonitorAttributes(ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "ClientGroup Monitor"



from typing import List
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.processing_engines import ProcessingEngines


class _ProcessingEngines(API):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/ProcessingEngines')

        self.Engine = self._Engine(api_obj=api_obj)
        self.Folder = self._Folder(api_obj=api_obj)

    def get(self):
        class _Response(APIResponse):
            def __init__(self, response):
                super().__init__(response=response)

            @property
            @api_response_property()
            def engines(self):
                return [ProcessingEngines.Engine(engine) for engine in self._from_json('Engines')]

        return _Response(response=self._get())

    class _Engine:
        def __init__(self, api_obj):
            self._api_obj = api_obj

        def Guid(self, guid: str):
            return self._Guid(guid=guid, api_obj=self._api_obj)

        class _Guid(API):
            def __init__(self, guid: str, api_obj):
                super().__init__(api_obj=api_obj, url=f'/ProcessingEngines/Engine/{guid}')

            def get(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property(return_on_204=list)
                    def folders(self):
                        return [ProcessingEngines.Folder(folder) for folder in self._from_json('Folders')][0]

                return _Response(response=self._get())

            def post(self, folder_guids: list):
                body = {
                    'FolderGuids': folder_guids
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response) \

                    @property
                    @api_response_property(return_on_204=str)
                    def added_count(self) -> int:
                        return self._from_json('AddedCount')

                    @property
                    @api_response_property(return_on_204=list)
                    def errors(self) -> List[str]:
                        return self._from_json('Errors')

                return _Response(response=self._post(data=body))

    class _Folder:
        def __init__(self, api_obj):
            self._api_obj = api_obj

        def Guid(self, guid: str):
            return self._Guid(guid=guid, api_obj=self._api_obj)

        class _Guid(API):
            def __init__(self, guid: str, api_obj):
                super().__init__(api_obj=api_obj, url=f'/ProcessingEngines/Folder/{guid}')
                self._folder_guid = guid

            def delete(self):
                return APIResponse(response=self._delete())

            def get(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property(return_on_204=list)
                    def engines(self):
                        return [ProcessingEngines.Engine(engine) for engine in self._from_json('Engines')]

                return _Response(response=self._get())

            def put(self, engine_guids: list):
                body = {
                    'EngineGuids': engine_guids
                }

                return APIResponse(response=self._put(data=body))

            def EngineGuid(self, guid):
                return self._EngineGuid(guid=self._folder_guid, engine_guid=guid, api_obj=self._api_obj)

            class _EngineGuid(API):
                def __init__(self, guid: str, engine_guid: str, api_obj):
                    super().__init__(api_obj=api_obj, url=f'/ProcessingEngines/Folder/{guid}/{engine_guid}')

                def delete(self):
                    return APIResponse(response=self._delete())



from pytpp.attributes._helper import IterableMeta, Attribute


class ProxyAttributes(metaclass=IterableMeta):
	__config_class__ = "Proxy"
	bypass_proxy_on_local = Attribute('Bypass Proxy on Local')
	credential = Attribute('Credential')
	proxy_credential = Attribute('Proxy Credential')
	proxy_host = Attribute('Proxy Host')
	proxy_port = Attribute('Proxy Port')
	proxy_use_host_configuration = Attribute('Proxy Use Host Configuration')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.metadata_base import MetadataBaseAttributes


class MetadataDateTimeAttributes(MetadataBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Metadata DateTime"
	date_only = Attribute('Date Only')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.validation_base import ValidationBaseAttributes


class X509CertificateValidationAttributes(ValidationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "X509 Certificate Validation"
	detect_all_ssl_tls_protocols = Attribute('Detect All SSL TLS Protocols', min_version='15.3')
	network_validation_disabled = Attribute('Network Validation Disabled', min_version='15.3')
	use_common_name = Attribute('Use Common Name', min_version='15.3')
	use_dns_subjectaltname = Attribute('Use DNS SubjectAltName', min_version='15.3')
	validate_chain_returned_by_host = Attribute('Validate Chain Returned By Host', min_version='15.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class CredentialRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Credential Root"
	escalation_notice_interval = Attribute('Escalation Notice Interval')
	escalation_notice_start = Attribute('Escalation Notice Start')
	expiration_notice_interval = Attribute('Expiration Notice Interval')
	expiration_notice_start = Attribute('Expiration Notice Start')
	protection_key = Attribute('Protection Key')
	start_time = Attribute('Start Time')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.flow_action_base import FlowActionBaseAttributes


class CodeSigningDeleteObjectActionAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Delete Object Action"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.report_base import ReportBaseAttributes


class ReportAssessorAttributes(ReportBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Report:Assessor"
	jobdn = Attribute('JobDN')
	options = Attribute('Options')



from typing import List, Union
from pytpp.tools.vtypes import Config
from pytpp.properties.config import PlacementRulesAttributeValues
from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.attributes.layout_rule_base import LayoutRuleBaseAttributes


@feature('Placement Rule Condition')
class PlacementRuleCondition:
    @property
    def city(self):
        """Base condition on the certificate city value."""
        return self._Operators(field='Certificate.City')

    @property
    def common_name(self):
        """Base condition on the certificate common name value."""
        return self._Operators(field='Certificate.CN')

    @property
    def country(self):
        """Base condition on the certificate country value."""
        return self._Operators(field='Certificate.C')

    @property
    def domain_component(self):
        """Base condition on the certificate domain component value."""
        return self._Operators(field='Certificate.DC')

    @property
    def expired(self):
        """Base condition on the certificate expiration value."""
        return self._Operators(field='Certificate.Expired')

    @property
    def hostname(self):
        """Base condition on the discovered hostname value."""
        return self._Operators(field='Discovery.Hostname')

    @property
    def ip_address(self):
        """Base condition on the discovered IP Address value."""
        return self._Operators(field='Discovery.Address')

    @property
    def issuer_dn(self):
        """Base condition on the certificate issuer :ref:`dn` value."""
        return self._Operators(field='Certificate.Issuer')

    @property
    def operating_system(self):
        """Base condition on the discovered Operating System value."""
        return self._Operators(field='Discovery.OS')

    @property
    def organization(self):
        """Base condition on the certificate organization value."""
        return self._Operators(field='Certificate.O')

    @property
    def organizational_unit(self):
        """Base condition on the certificate organizational unit value."""
        return self._Operators(field='Certificate.OU')

    @property
    def port(self):
        """Base condition on the discovered port value."""
        return self._Operators(field='Discovery.Port')

    @property
    def san(self):
        """Base condition on the certificate SAN DNS value."""
        return self._Operators(field='Certificate.SANDNS')

    @property
    def self_signed(self):
        """Base condition on whether or not the certificate is self-signed."""
        return self._Operators(field='Certificate.SelfSigned')

    @property
    def server_version(self):
        """Base condition on the SSH server version value."""
        return self._Operators(field='SSH.ServerVersion')

    @property
    def state(self):
        """Base condition on the certificate state value."""
        return self._Operators(field='Certificate.State')

    @property
    def supports_ssh_v1(self):
        """Base condition on if the target supports SSH V1."""
        return self._Operators(field='SSH.Version', force_value='1')

    @property
    def supports_ssh_v2(self):
        """Base condition on if the target supports SSH V2."""
        return self._Operators(field='SSH.Version', force_value='2')

    class _Operators:
        def __init__(self, field: str, force_value: str = None):
            self._field = field
            self._forced_value = force_value

        def matches(self, value: str):
            """
            Args:
                value: String to match.

            Returns: Condition string for the Placement Rule to be created.
            """
            return f'{self._field} == "{value}"'

        def in_list(self, values: List[str]):
            """
            Args:
                values: List of strings to match.

            Returns: Condition string for the Placement Rule to be created.
            """
            return f"{self._field} in [{','.join(map(str, values))}]"

        def starts_with(self, value: str):
            """
            Args:
                value: String that the target field's value begins with.

            Returns: Condition string for the Placement Rule to be created.
            """
            return f'{self._field} > "{value}"'

        def ends_with(self, value: str):
            """
            Args:
                value: String that the target field's value ends with.

            Returns: Condition string for the Placement Rule to be created.
            """
            return f'{self._field} < "{value}"'

        def contains(self, value: str):
            """
            Args:
                value: String within the target field's value.

            Returns: Condition string for the Placement Rule to be created.
            """
            return f'{self._field} like "{value}"'

        def matches_regex(self, value: str):
            """
            Args:
                value: String to match by regular expression.

            Returns: Condition string for the Placement Rule to be created.
            """
            return f'{self._field} @= "{value}"'

        def is_true(self):
            """
            Returns: Condition string for the Placement Rule to be created.
            """
            if self._forced_value:
                return f"{self._field} == {self._forced_value}"
            return f"{self._field} == 1"

        def is_false(self):
            """
            Returns: Condition string for the Placement Rule to be created.
            """
            if self._forced_value:
                return f"{self._field} != {self._forced_value}"
            return f"{self._field} == 0"


@feature('Placement Rules')
class PlacementRules(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._layout_rules_dn = r'\VED\Layout Root\Rules'

    def _format_rule_attribute(self, conditions: List[str], device_location: 'Union[Config.Object, str]',
                               certificate_location: 'Union[Config.Object, str]' = None,
                               rule_type: str = 'X509 Certificate'):
        """
        Formats the rule attribute on the Placement Rule object.
        """
        context = 'CONTEXT\n\tDiscovery'
        rule_types = f"FOR\n\t{','.join(('Device', rule_type))}"
        conditions = f"IF\n\t{' && '.join(conditions)}"
        locations = [f'Location[Device]={self._get_dn(device_location)}']
        if certificate_location:
            locations.append(f'Location[X509 Certificate Base]={self._get_dn(certificate_location)}')
        locations = f"THEN\n\t" + '\n\t'.join(locations)

        return f"{context}\n{rule_types}\n{conditions}\n{locations}\nEND"

    def create(self, name: str, conditions: List[str], device_location: 'Union[Config.Object, str]',
               certificate_location: 'Union[Config.Object, str]' = None, rule_type: str = 'X509 Certificate',
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the placement rule.
            conditions: A list of strings that define the conditions of the rule. Use :class:`PlacementRuleCondition` to 
                        create the rule conditions. See :ref:`placement_usage` for examples.
            device_location: :ref:`config_object` or :ref:`dn` of the folder in which to place applicable devices.
            certificate_location: :ref:`config_object` or :ref:`dn` of the folder in which to place applicable certificates.
            rule_type: Default is 'X509 Certificate'. 'SSH' may be specified instead for SSH discovery.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the placement rule object.
        """
        rule_attr = self._format_rule_attribute(
            conditions=conditions,
            device_location=device_location,
            certificate_location=certificate_location,
            rule_type=rule_type
        )
        rule = self._config_create(
            name=name,
            parent_folder_dn=self._layout_rules_dn,
            attributes={
                LayoutRuleBaseAttributes.rule: rule_attr
            },
            config_class=LayoutRuleBaseAttributes.__config_class__,
            get_if_already_exists=get_if_already_exists
        )
        return rule

    def delete(self, rule: 'Union[Config.Object, str]'):
        """
        Deletes a placement rule.

        Args:
            rule: :ref:`config_object` or name of the placement rule.
        """
        rule_dn = self._get_dn(rule, parent_dn=self._layout_rules_dn)
        response = self._config_delete(object_dn=rule_dn)
        response.assert_valid_response()

    def update(self, rule: 'Union[Config.Object, str]', conditions: List[str] = None, device_location: str = None,
               certificate_location: str = None, rule_type: str = 'X509 Certificate'):
        """
        Updates a placement rule. If certain parameters are not provided, the current parameters will be rewritten
        to the object. In other words, only the parameters given are updated.

        Args:
            rule: :ref:`config_object` or name of the placement rule.
            conditions: A list of strings that define the conditions of the rule. Use :class:`PlacementRuleCondition` to
                        create the rule conditions. See :ref:`placement_usage` for examples.
            device_location: :ref:`config_object` or :ref:`dn` of the folder in which to place applicable devices.
            certificate_location: :ref:`config_object` or :ref:`dn` of the folder in which to place applicable certificates.
            rule_type: Default is 'X509 Certificate'. 'SSH' may be specified instead for SSH discovery.
        """
        rule_dn = self._get_dn(rule, parent_dn=self._layout_rules_dn)
        current_attr = self._api.websdk.Config.Read.post(
            object_dn=rule_dn,
            attribute_name=LayoutRuleBaseAttributes.rule
        ).values[0]
        new_conditions = conditions or []
        new_certificate_dn = self._get_dn(certificate_location) if certificate_location else ''
        new_device_dn = self._get_dn(device_location) if device_location else ''
        get_conditions = get_locations = False
        for line in current_attr.splitlines():
            if line.strip() == 'IF':
                if not conditions:
                    get_conditions = True
                get_locations = False
            elif line.strip() == 'THEN':
                get_conditions = False
                if not(new_certificate_dn and new_device_dn):
                    get_locations = True
            elif line.strip() == 'END':
                break
            elif get_conditions:
                new_conditions = line.strip().split(' && ')
                get_conditions = False
            elif get_locations:
                if 'Location[Device]' in line and not new_device_dn:
                    new_device_dn = line.split('=', 1)[-1].strip()
                elif 'Location[X509 Certificate Base]' in line and not new_certificate_dn:
                    new_certificate_dn = line.split('=', 1)[-1].strip()

        if rule_type == PlacementRulesAttributeValues.RuleType.ssh:
            new_certificate_dn = None

        rule_attr = self._format_rule_attribute(
            conditions=new_conditions,
            device_location=new_device_dn,
            certificate_location=new_certificate_dn,
            rule_type=rule_type
        )
        response = self._api.websdk.Config.WriteDn.post(
            object_dn=rule_dn,
            attribute_name=LayoutRuleBaseAttributes.rule,
            values=[
                rule_attr
            ]
        )
        response.assert_valid_response()

    def get(self, name: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            name: Name of the placement rule.
            raise_error_if_not_exists: Raise an exception if the placement rule does not exist.

        Returns:
            :ref:`config_object` of the placement rule object.
        """
        return self._get_config_object(
            object_dn=f'{self._layout_rules_dn}\\{name}',
            raise_error_if_not_exists=raise_error_if_not_exists
        )



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.schedule_base import ScheduleBaseAttributes
from pytpp.attributes.service_module import ServiceModuleAttributes


class CloudInstanceMonitorAttributes(ScheduleBaseAttributes, ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "Cloud Instance Monitor"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class iPlanetAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "iPlanet"
	alias = Attribute('Alias')
	certutil_path = Attribute('Certutil Path')
	create_store = Attribute('Create Store')
	create_virtual_server = Attribute('Create Virtual Server')
	database_credential = Attribute('Database Credential')
	database_prefix = Attribute('Database Prefix')
	database_type = Attribute('Database Type', min_version='19.1')
	database_validation_disabled = Attribute('Database Validation Disabled')
	document_root = Attribute('Document Root')
	file_validation_disabled = Attribute('File Validation Disabled', min_version='15.3')
	install_path = Attribute('Install Path')
	key_store = Attribute('Key Store')
	key_store_credential = Attribute('Key Store Credential')
	mta_host = Attribute('MTA Host')
	network_validation_disabled = Attribute('Network Validation Disabled')
	pk12util_path = Attribute('Pk12util Path')
	protocol = Attribute('Protocol')
	replace_store = Attribute('Replace Store')
	secure_server_name = Attribute('Secure Server Name')
	use_proxy = Attribute('Use Proxy')
	virtual_server_dns_value = Attribute('Virtual Server DNS Value')
	virtual_server_port = Attribute('Virtual Server Port')
	virtual_server_user = Attribute('Virtual Server User')
	web_credential = Attribute('Web Credential')
	web_port = Attribute('Web Port')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class VamCaviumAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "VamCavium"
	cavium_utility_path = Attribute('Cavium Utility Path')
	key_list_path = Attribute('Key List Path')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class WorkflowRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Workflow Root"
	approval_reason = Attribute('Approval Reason')
	approver_not_found_expiration = Attribute('Approver Not Found Expiration', min_version='16.2')
	expiration_days = Attribute('Expiration Days')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.monitoring_base import MonitoringBaseAttributes
from pytpp.attributes.top import TopAttributes


class KeyBaseAttributes(MonitoringBaseAttributes, TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Key Base"
	algorithm = Attribute('Algorithm', min_version='20.2')
	approver = Attribute('Approver', min_version='20.2')
	consumers = Attribute('Consumers', min_version='20.2')
	disable_automatic_renewal = Attribute('Disable Automatic Renewal', min_version='20.2')
	in_error = Attribute('In Error', min_version='20.2')
	internet_email_address = Attribute('Internet EMail Address', min_version='20.2')
	key_storage_location = Attribute('Key Storage Location', min_version='20.2')
	label = Attribute('Label', min_version='20.2')
	last_evaluated_on = Attribute('Last Evaluated On', min_version='20.2')
	last_notification = Attribute('Last Notification', min_version='20.2')
	last_renewed_by = Attribute('Last Renewed By', min_version='20.2')
	last_renewed_on = Attribute('Last Renewed On', min_version='20.2')
	management_type = Attribute('Management Type', min_version='20.2')
	max_uses = Attribute('Max Uses', min_version='20.2')
	origin = Attribute('Origin', min_version='20.2')
	private_key_vault_id = Attribute('Private Key Vault Id', min_version='20.2')
	real_name = Attribute('Real Name', min_version='20.2')
	status = Attribute('Status', min_version='20.2')
	use_count = Attribute('Use Count', min_version='20.2')
	validity_period = Attribute('Validity Period', min_version='20.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.key_base import KeyBaseAttributes


class KeyPairAttributes(KeyBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Key Pair"
	created_by = Attribute('Created By', min_version='20.3')
	gpg_public_key_vault_id = Attribute('GPG Public Key Vault ID', min_version='21.4')
	key_pair_renewal_flow = Attribute('Key Pair Renewal Flow', min_version='20.2')
	original_creation_date = Attribute('Original Creation Date', min_version='20.2')
	public_key_vault_id = Attribute('Public Key Vault Id', min_version='20.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.flow_action_approver_base import FlowActionApproverBaseAttributes


class CodeSigningAdminApprovalActionAttributes(FlowActionApproverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Admin Approval Action"
	include_code_signing_admin = Attribute('Include Code Signing Admin', min_version='21.1')
	include_master_admin = Attribute('Include Master Admin', min_version='21.1')



from dataclasses import dataclass
from datetime import datetime
from typing import List


@dataclass
class Result:
    code: int
    flow_result: str


@dataclass
class Ticket:
    approvals: 'List[Approval]'
    approvers: List[str]
    creation_time: datetime
    environment: 'List[KeyValue]'
    flow_process_id: int
    id: int
    identifier: str
    product_code: int
    remaining_uses: int
    required_approvals: int


@dataclass
class Approval:
    approval_time: datetime
    universal: str


@dataclass
class KeyValue:
    key: str
    value: str



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class VamSunimpAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "VamSunimp"
	file_validation_disabled = Attribute('File Validation Disabled', min_version='15.3')
	key_list_path = Attribute('Key List Path')
	key_store_credential = Attribute('Key Store Credential')
	key_store_name = Attribute('Key Store Name')
	key_store_validation_disabled = Attribute('Key Store Validation Disabled')
	sca6000_utility_path = Attribute('SCA6000 Utility Path')
	sunimp_utility_path = Attribute('Sunimp Utility Path')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class CodeSigningApplicationCollectionAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Application Collection"
	code_signing_application_dn = Attribute('Code Signing Application DN', min_version='19.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class JKSAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "JKS"
	certificate_label = Attribute('Certificate Label')
	create_store = Attribute('Create Store')
	disable_ssh_history = Attribute('Disable SSH History', min_version='19.3')
	file_validation_disabled = Attribute('File Validation Disabled', min_version='15.3')
	java_vendor = Attribute('Java Vendor', min_version='18.1')
	key_algorithm = Attribute('Key Algorithm')
	key_store = Attribute('Key Store')
	key_store_credential = Attribute('Key Store Credential')
	key_store_validation_disabled = Attribute('Key Store Validation Disabled')
	keytool_path = Attribute('Keytool Path')
	network_validation_disabled = Attribute('Network Validation Disabled')
	private_key_location = Attribute('Private Key Location', min_version='17.2')
	protection_type = Attribute('Protection Type', min_version='18.1')
	recycle_alias = Attribute('Recycle Alias')
	replace_store = Attribute('Replace Store')
	slot_number = Attribute('Slot Number', min_version='17.2')
	softcard_identifier = Attribute('Softcard Identifier', min_version='18.1')
	store_type = Attribute('Store Type')
	temp_certificate_label = Attribute('Temp Certificate Label')
	version = Attribute('Version')



from typing import List
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.client import Client


class _Client(API):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Client')
        self.Delete = self._Delete(api_obj=api_obj)
        self.Details = self._Details(api_obj=api_obj)
        self.Work = self._Work(api_obj=api_obj)

    def get(self, client_version: int = None, client_type: str = None, host_name: str = None, ip_address: str = None,
            last_seen_on: str = None,last_seen_on_greater: str = None, last_seen_on_less: str = None, mac_address: str = None,
            os_name: str = None, os_version: str = None, region: str = None, serial_number: int = None, sid: str = None,
            user_name: str = None, virtual_machine_id: int = None):
        params = {
            'ClientVersion': client_version,
            'ClientType': client_type,
            'HostName': host_name,
            'IpAddress': ip_address,
            'LastSeenOn': last_seen_on,
            'LastSeenOnGreater': last_seen_on_greater,
            'LastSeenOnLess': last_seen_on_less,
            'MacAddress': mac_address,
            'OSName': os_name,
            'OSVersion': os_version,
            'Region': region,
            'SerialNumber': serial_number,
            'SID': sid,
            'UserName': user_name,
            'VirtualMachineId': virtual_machine_id
        }

        class _Response(APIResponse):
            def __init__(self, response):
                super().__init__(response=response)

            @property
            @api_response_property(return_on_204=list)
            def clients(self):
                return [Client.Client(client) for client in self._from_json()]

        return _Response(response=self._get(params=params))

    class _Delete(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Client/Delete')

        def post(self, clients: list, delete_associated_devices: bool = False):
            body = {
                'Clients': clients,
                'DeleteAssociatedDevices': delete_associated_devices
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def deleted_count(self) -> int:
                    return self._from_json(key='DeletedCount')

                @property
                @api_response_property()
                def errors(self) -> List[str]:
                    return self._from_json(key='Errors')

            return _Response(response=self._post(data=body))

    class _Details(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Client/Details')

        def get(self, client_version: int = None, client_type: str = None, host_name: str = None, ip_address: str = None,
                last_seen_on: str = None, last_seen_on_greater: str = None, last_seen_on_less: str = None, mac_address: str = None,
                os_name: str = None, os_version: str = None, region: str = None, serial_number: int = None, sid: str = None, user_name: str = None,
                virtual_machine_id: int = None):
            params = {
                'ClientVersion': client_version,
                'ClientType': client_type,
                'HostName': host_name,
                'IpAddress': ip_address,
                'LastSeenOn': last_seen_on,
                'LastSeenOnGreater': last_seen_on_greater,
                'LastSeenOnLess': last_seen_on_less,
                'MacAddress': mac_address,
                'OSName': os_name,
                'OSVersion': os_version,
                'Region': region,
                'SerialNumber': serial_number,
                'SID': sid,
                'UserName': user_name,
                'VirtualMachineId': virtual_machine_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property(return_on_204=list)
                def details(self):
                    return [Client.ClientDetails(client) for client in self._from_json()]

            return _Response(response=self._get(params=params))

    class _Work(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Client/Work')

        def get(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property(return_on_204=list)
                def works(self):
                    return [Client.Work(work) for work in self._from_json()]

            return _Response(response=self._get())



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class LogChannelContainerAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Log Channel Container"
	description = Attribute('Description')



from typing import List 
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.identity import Identity


class _Identity:
    def __init__(self, api_obj):
        self.AddGroup = self._AddGroup(api_obj=api_obj)
        self.AddGroupMembers = self._AddGroupMembers(api_obj=api_obj)
        self.Browse = self._Browse(api_obj=api_obj)
        self.GetAssociatedEntries = self._GetAssociatedEntries(api_obj=api_obj)
        self.GetMembers = self._GetMembers(api_obj=api_obj)
        self.GetMemberships = self._GetMemberships(api_obj=api_obj)
        self.Group = self._Group(api_obj=api_obj)
        self.ReadAttribute = self._ReadAttribute(api_obj=api_obj)
        self.RemoveGroupMembers = self._RemoveGroupMembers(api_obj=api_obj)
        self.RenameGroup = self._RenameGroup(api_obj=api_obj)
        self.Self = self._Self(api_obj=api_obj)
        self.SetPassword = self._SetPassword(api_obj=api_obj)
        self.Validate = self._Validate(api_obj=api_obj)

    class _AddGroup(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/AddGroup')

        def post(self, name: str, members: list = None):
            body = {
                'Name': name,
                'Members': members
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def identity(self):
                    return Identity.Identity(self._from_json('ID'))

                @property
                @api_response_property()
                def invalid_members(self):
                    return [Identity.InvalidIdentity(im) for im in self._from_json(key='InvalidMembers', return_on_error=list)]

                @property
                @api_response_property()
                def message(self) -> str:
                    return self._from_json(key='Message', return_on_error=str)

            return _Response(response=self._post(data=body))

    class _AddGroupMembers(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/AddGroupMembers')

        def put(self, group: dict, members: list, show_members: bool = False):
            body = {
                'Group': group,
                'Members': members,
                'ShowMembers': show_members
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def invalid_members(self):
                    return [Identity.InvalidIdentity(im) for im in self._from_json('InvalidMembers', return_on_error=list)]

                @property
                @api_response_property()
                def members(self):
                    return [Identity.Identity(m) for m in self._from_json('Members', return_on_error=list)]

                @property
                @api_response_property()
                def message(self) -> str:
                    return self._from_json('Message')

            return _Response(response=self._put(data=body))

    class _Browse(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/Browse')

        def post(self, filter: str, limit: int, identity_type: int):
            body = {
                "Filter": filter,
                "Limit": limit,
                "IdentityType": identity_type
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def identities(self):
                    return [Identity.Identity(i) for i in self._from_json(key='Identities', return_on_error=list)]

            return _Response(response=self._post(data=body))

    class _GetAssociatedEntries(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/GetAssociatedEntries')

        def post(self, identity: dict):
            body = {
                'ID': identity
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def identities(self):
                    return [Identity.Identity(i) for i in self._from_json(key='Identities', return_on_error=list)]

            return _Response(response=self._post(data=body))

    class _GetMembers(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/GetMembers')

        def post(self, identity: dict, resolve_nested: bool = False):
            body = {
                'ID': identity,
                'ResolveNested': int(resolve_nested)
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def identities(self):
                    return [Identity.Identity(i) for i in self._from_json(key='Identities', return_on_error=list)]

            return _Response(response=self._post(data=body))

    class _GetMemberships(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/GetMemberships')

        def post(self, identity: dict):
            body = {
                'ID': identity
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def identities(self):
                    return [Identity.Identity(i) for i in self._from_json(key='Identities', return_on_error=list)]

            return _Response(response=self._post(data=body))

    class _Group:
        def __init__(self, api_obj):
            self._api_obj = api_obj

        def Prefix(self, prefix: str):
            return self._Prefix(prefix=prefix, api_obj=self._api_obj)

        class _Prefix:
            def __init__(self, prefix: str, api_obj):
                self._prefix = prefix
                self._api_obj = api_obj

            def Principal(self, principal: str):
                return self._Principal(prefix=self._prefix, principal=principal, api_obj=self._api_obj)

            class _Principal(API):
                def __init__(self, prefix: str, principal: str, api_obj):
                    super().__init__(api_obj=api_obj, url=f'/Identity/Group/{prefix}/{principal}')

                def delete(self):
                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                        @property
                        @api_response_property()
                        def message(self) -> str:
                            return self._from_json('Message')

                    return _Response(response=self._delete())

    class _ReadAttribute(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/ReadAttribute')

        def post(self, attribute_name: str, identity: dict):
            body = {
                'ID': identity,
                'AttributeName': attribute_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def attributes(self) -> List[str]:
                    return self._from_json(key='Attributes')

            return _Response(response=self._post(data=body))

    class _RemoveGroupMembers(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/RemoveGroupMembers')

        def put(self, group: dict, members: list, show_members: bool = False):
            body = {
                'Group': group,
                'Members': members,
                'ShowMembers': show_members
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def invalid_members(self):
                    return [Identity.InvalidIdentity(im) for im in self._from_json('InvalidMembers', return_on_error=list)]

                @property
                @api_response_property()
                def members(self):
                    return [Identity.Identity(m) for m in self._from_json('Members', return_on_error=list)]

                @property
                @api_response_property()
                def message(self) -> str:
                    return self._from_json('Message')

            return _Response(response=self._put(data=body))

    class _RenameGroup(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/RenameGroup')

        def put(self, group: dict, new_group_name: str):
            body = {
                'Group': group,
                'NewGroupName': new_group_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def identity(self):
                    return Identity.Identity(self._from_json('ID'))

            return _Response(response=self._put(data=body))

    class _Self(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/Self')

        def get(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def identities(self):
                    return [Identity.Identity(i) for i in self._from_json(key='Identities')]

            return _Response(response=self._get())

    class _SetPassword(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/SetPassword')

        def post(self, identity: dict, password: str, old_password: str = None):
            body = {
                'ID': identity,
                'OldPassword': old_password,
                'Password': password
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def identity(self):
                    return Identity.Identity(self._from_json('ID'))

            return _Response(response=self._post(data=body))

    class _Validate(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Identity/Validate')

        def post(self, identity: dict):
            body = {
                'ID': identity
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def identity(self):
                    return Identity.Identity(self._from_json('ID'))

            return _Response(response=self._post(data=body))



from pytpp.attributes.microsoft_ca import MicrosoftCAAttributes
from pytpp.attributes.self_signed_ca import SelfSignedCAAttributes
from pytpp.features.bases.feature_base import FeatureBase, feature
from typing import Union, List, TYPE_CHECKING
if TYPE_CHECKING:
    from pytpp.tools.vtypes import Config, Identity


class _CertificateAuthorityBase(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)

    def delete(self, certificate_authority: 'Union[Config.Object, str]'):
        """
        Deletes the certificate authority object from TPP, including all of the secrets associated to it.

        Args:
            certificate_authority: :ref:`config_object` or :ref:`dn` for the certificate authority object.
        """
        certificate_authority_dn = self._get_dn(certificate_authority)
        self._secret_store_delete(object_dn=certificate_authority_dn)
        self._config_delete(object_dn=certificate_authority_dn)

    def get(self, certificate_authority_dn: str, raise_error_if_not_exists: bool = True):
        """
        Get the certificate authority object in TPP.

        Args:
            certificate_authority_dn: :ref:`dn` of the certificate authority.
            raise_error_if_not_exists: Raise an exception if the object :ref:`dn` does not exist.

        Returns:
            :ref:`config_object` of the certificate authority.
        """
        return self._get_config_object(
            object_dn=certificate_authority_dn,
            raise_error_if_not_exists=raise_error_if_not_exists
        )


@feature('Microsoft CA')
class MSCA(_CertificateAuthorityBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', hostname: str, service_name: str,
               credential: 'Union[Config.Object, str]', template: str, description: 'str' = None,
               contacts: 'List[Identity.Identity, str]' = None, manual_approvals: 'bool' = None,
               subject_alt_name_enabled: 'bool' = None, automatically_include_cn_as_dns_san: 'bool' = None,
               allow_users_to_specify_end_date: 'bool' = None, enrollment_agent: 'Union[Config.Object, str]' = None,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the CA object.
            parent_folder: `:ref:`config_object` or :ref:`dn` of the parent folder of this certificate authority object.
            hostname: Hostname or IP Address of the CA.
            service_name: Service, or Given, Name of the certificate authority.
            credential: `:ref:`config_object` or :ref:`dn` of the CA creential.
            template: Name of the CA template.
            description: Description of the CA object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` of the contacts.
            manual_approvals: Require manual approvals.
            subject_alt_name_enabled: Enable Subject Alterntaive Names.
            automatically_include_cn_as_dns_san: Automatically include the common name (CN) as a DNS SAN.
            allow_users_to_specify_end_date: Allow users to specify the end date.
            enrollment_agent: `:ref:`config_object` or :ref:`dn` of the certificate credential, or enrollment agent.
            attributes: Additional attributes associated to the CA object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the certificate authority.
        """
        ca_attrs = {
            MicrosoftCAAttributes.driver_name: 'camicrosoft',
            MicrosoftCAAttributes.host: hostname,
            MicrosoftCAAttributes.given_name: service_name,
            MicrosoftCAAttributes.credential: self._get_dn(credential),
            MicrosoftCAAttributes.template: template,
            MicrosoftCAAttributes.description: description,
            MicrosoftCAAttributes.contact: [self._get_prefixed_universal(c) for c in contacts] if contacts else None,
            MicrosoftCAAttributes.manual_approval: {True: "1", False: "0"}.get(manual_approvals),
            MicrosoftCAAttributes.san_enabled: {True: "1", False: "1"}.get(subject_alt_name_enabled),
            MicrosoftCAAttributes.include_cn_as_san: {True: "1", False: "0"}.get(automatically_include_cn_as_dns_san),
            MicrosoftCAAttributes.specific_end_date_enabled: {True: "1", False: "0"}.get(allow_users_to_specify_end_date),
            MicrosoftCAAttributes.enrollment_agent_certificate: self._get_dn(enrollment_agent) if enrollment_agent else None
        }
        if attributes:
            ca_attrs.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._get_dn(parent_folder),
            config_class=MicrosoftCAAttributes.__config_class__,
            attributes=ca_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Self-Signed CA')
class SelfSignedCA(_CertificateAuthorityBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Identity.Identity, str]' = None, key_usage: 'List[str]' = None, server_authentication: 'bool' = None,
               client_authentication: 'bool' = None, code_signing: 'bool' = None, signature_algorithm: 'str' = None,
               valid_years: 'int' = None, valid_days: 'int' = None, attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the CA object.
            parent_folder: `:ref:`config_object` or :ref:`dn` of the parent folder of this certificate authority object.
            description: Description of the CA object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` of the contacts.
            key_usage: List of key usages.
            server_authentication: Allow server authentication.
            client_authentication: Allow client authentication.
            code_signing: Allow code signing.
            signature_algorithm: Signing algorithm.
            valid_years: Validity period in years.
            valid_days: Validitiy period in days. Added to years.
            attributes: Additional attributes associated to the CA object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the certificate authority.
        """
        ca_attrs = {
            SelfSignedCAAttributes.driver_name: 'caselfsigned',
            SelfSignedCAAttributes.description: description,
            SelfSignedCAAttributes.contact: [self._get_prefixed_universal(c) for c in contacts] if contacts else None,
            SelfSignedCAAttributes.key_usage: ','.join(key_usage),
            SelfSignedCAAttributes.algorithm: signature_algorithm,
        }
        if server_authentication or client_authentication or code_signing:
            enhanced_key_usage = {
                '1.3.6.1.5.5.7.3.1': server_authentication,
                '1.3.6.1.5.5.7.3.2': client_authentication,
                '1.3.6.1.5.5.7.3.3': code_signing
            }
            ca_attrs.update({
                SelfSignedCAAttributes.enhanced_key_usage: [
                    eku for eku, enabled in enhanced_key_usage.items() if enabled is True
                ],
            })
        if valid_years or valid_days:
            validity_period= (365 * (valid_years or 0)) + (valid_days or 0)
            ca_attrs.update({
                SelfSignedCAAttributes.validity_period: validity_period
            })
        if attributes:
            ca_attrs.update(attributes)
        return self._config_create(
            name=name,
            parent_folder_dn=self._get_dn(parent_folder),
            config_class=SelfSignedCAAttributes.__config_class__,
            attributes=ca_attrs,
            get_if_already_exists=get_if_already_exists
        )



from pytpp.properties.response_objects.dataclasses import ssh
from pytpp.properties.resultcodes import ResultCodes
from pytpp.tools.helpers.date_converter import from_date_string


class SSH:
    @staticmethod
    def Response(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        error_code = response_object.get('ErrorCode')
        return ssh.Response(
            success=response_object.get('Success'),
            error_code=error_code,
            error_message=ResultCodes.SSHErrorCodes.get(error_code, 'Unknown') if error_code else None,
        )

    @staticmethod
    def ConnectionResult(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return ssh.ConnectionResult(
            device_guid=response_object.get('DeviceGuid'),
            error=response_object.get('Error'),
            success=response_object.get('Success'),
        )

    @staticmethod
    def DeviceData(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return ssh.DeviceData(
            dn=response_object.get('DN'),
            device_guid=response_object.get('DeviceGuid'),
            host_name=response_object.get('HostName'),
            is_compliant=response_object.get('IsCompliant'),
            type=response_object.get('Type'),
        )

    @staticmethod
    def KeyData(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return ssh.KeyData(
            active_from=from_date_string(response_object.get('ActiveFrom')),
            algorithm=response_object.get('Algorithm'),
            allowed_source_restriction=response_object.get('AllowedSourceRestriction'),
            approver=response_object.get('Approver'),
            comment=response_object.get('Comment'),
            denied_source_restriction=response_object.get('DeniedSourceRestriction'),
            device_guid=response_object.get('DeviceGuid'),
            filepath=response_object.get('Filepath'),
            forced_command=response_object.get('ForcedCommand'),
            format=response_object.get('Format'),
            is_encrypted=response_object.get('IsEncrypted'),
            key_id=response_object.get('KeyId'),
            keysetid=response_object.get('KeysetId'),
            last_used=from_date_string(response_object.get('LastUsed')),
            length=response_object.get('Length'),
            notes=response_object.get('Notes'),
            options=response_object.get('Options'),
            process_error=response_object.get('ProcessError'),
            process_status=response_object.get('ProcessStatus'),
            reason=response_object.get('Reason'),
            rotation_stage=response_object.get('RotationStage'),
            type=response_object.get('Type'),
            username=response_object.get('Username'),
            violation_status=response_object.get('ViolationStatus'),
        )

    @staticmethod
    def KeySetData(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return ssh.KeySetData(
            access=response_object.get('Access'),
            algorithm=response_object.get('Algorithm'),
            fingerprint_md5=response_object.get('FingerprintMD5'),
            fingerprint_sha256=response_object.get('FingerprintSHA256'),
            keysetid=response_object.get('KeysetId'),
            last_rotation_date=from_date_string(response_object.get('LastRotationDate')),
            last_used=from_date_string(response_object.get('LastUsed')),
            length=response_object.get('Length'),
            private_keys=[SSH.KeyData(data) for data in
                          response_object.get('PrivateKeys')] if 'PrivateKeys' in response_object.keys() else [],
            process_error=response_object.get('ProcessError'),
            process_status=response_object.get('ProcessStatus'),
            public_keys=[SSH.KeyData(data) for data in
                         response_object.get('PublicKeys')] if 'PublicKeys' in response_object.keys() else [],
            rotation_stage=response_object.get('RotationStage'),
            type=response_object.get('Type'),
            violation_status=response_object.get('ViolationStatus'),
        )

    @staticmethod
    def KeyUsageData(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return ssh.KeyUsageData(
            alert=response_object.get('Alert'),
            authorized_key_id=response_object.get('AuthorizedKeyId'),
            client_name=response_object.get('ClientName'),
            fingerprint=response_object.get('Fingerprint'),
            key_usage_id=response_object.get('KeyUsageId'),
            keyset_id=response_object.get('KeysetId'),
            last_used=response_object.get('LastUsed'),
            private_key_id=response_object.get('PrivateKeyId'),
            server_account=response_object.get('ServerAccount'),
            server_name=response_object.get('ServerName'),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class GoogleCloudAppAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Google Cloud App"
	file_validation_disabled = Attribute('File Validation Disabled', min_version='20.2')
	network_validation_disabled = Attribute('Network Validation Disabled', min_version='20.2')
	target_proxy_name = Attribute('Target Proxy Name', min_version='20.2')
	target_proxy_type = Attribute('Target Proxy Type', min_version='20.2')
	target_region = Attribute('Target Region', min_version='21.4')
	target_resource = Attribute('Target Resource', min_version='20.2')
	timeout = Attribute('Timeout', min_version='20.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes


class SelfSignedCAAttributes(CertificateAuthorityBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Self Signed CA"
	algorithm = Attribute('Algorithm')
	enhanced_key_usage = Attribute('Enhanced Key Usage')
	key_usage = Attribute('Key Usage')



class Namespaces:
    config = 'config'


class KeyNames:
    software_default = 'Software:Default'
    null_null = 'Null:Null'


class VaultTypes:
    none = 0
    private_key = 1
    certificate = 2
    pkcs12 = 4
    symmetric_key = 8
    state = 16  # Obsolete
    password = 32
    csr = 64  # Obsolete: use pkcs10 instead
    blob = 128
    pkcs8 = 256
    pkcs10 = 512
    file = 1024
    rsa_public_key = 2048  # Obsolete: use archived_public_key instead
    public_key = 4096
    archived_certificate = 1073741826
    archived_pkcs12 = 1073741828
    archived_symmetric_key = 1073741832
    archived_state = 1073741840  # Obsolete
    archived_password = 1073741856
    archived_blob = 1073741952
    archived_pkcs8 = 1073742080
    archived_pkcs10 = 1073742336
    archived_file = 1073742848
    archived_rsa_public_key  = 1073743872  # Obsolete: use archived_public_key instead
    archived_public_key = 1073745920



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes
from pytpp.attributes.proxy import ProxyAttributes


class AdaptableCAAttributes(CertificateAuthorityBaseAttributes, ProxyAttributes, metaclass=IterableMeta):
	__config_class__ = "Adaptable CA"
	allow_reissue = Attribute('Allow Reissue', min_version='16.2')
	certificate_credential = Attribute('Certificate Credential', min_version='16.2')
	connection_valid = Attribute('Connection Valid', min_version='16.2')
	custom_fields = Attribute('Custom Fields', min_version='16.2')
	interoperability_script = Attribute('Interoperability Script', min_version='16.2')
	log_debug = Attribute('Log Debug', min_version='19.3')
	powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id', min_version='18.4')
	profile_string = Attribute('Profile String', min_version='20.1')
	renewal_window = Attribute('Renewal Window', min_version='16.2')
	retry_after_script_hash_mismatch = Attribute('Retry After Script Hash Mismatch', min_version='18.4')
	script_execution_timeout = Attribute('Script Execution Timeout', min_version='20.2')
	secondary_credential = Attribute('Secondary Credential', min_version='17.4')
	service_address = Attribute('Service Address', min_version='20.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.exclusion import ExclusionAttributes


class DiscoveryExclusionAttributes(ExclusionAttributes, metaclass=IterableMeta):
	__config_class__ = "Discovery Exclusion"
	address_range = Attribute('Address Range')
	managed_by_discovery_dn = Attribute('Managed By Discovery DN')
	managed_dn = Attribute('Managed DN')
	port_range = Attribute('Port Range')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes
from pytpp.attributes.proxy import ProxyAttributes


class VeriSignCAAttributes(HTTPCABaseAttributes, ProxyAttributes, metaclass=IterableMeta):
	__config_class__ = "VeriSign CA"
	additional_field = Attribute('Additional Field')
	admin_isd = Attribute('Admin Isd')
	allow_reissue = Attribute('Allow Reissue', min_version='17.2')
	attempt_renewal = Attribute('Attempt Renewal', min_version='17.2')
	certificate_block = Attribute('Certificate Block')
	certificate_transparency = Attribute('Certificate Transparency', min_version='16.3')
	domain = Attribute('Domain')
	interval = Attribute('Interval')
	jurisdiction_hash = Attribute('Jurisdiction Hash')
	mpki_url = Attribute('Mpki Url')
	organization = Attribute('Organization')
	organizational_unit = Attribute('Organizational Unit')
	renewal_window = Attribute('Renewal Window', min_version='17.2')
	retrieval_period = Attribute('Retrieval Period')
	vice_url = Attribute('Vice Url')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.encryption_driver import EncryptionDriverAttributes


class NullEncryptionDriverAttributes(EncryptionDriverAttributes, metaclass=IterableMeta):
	__config_class__ = "Null Encryption Driver"



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.log_sql_channel import LogSQLChannelAttributes


class LogMSSQLAttributes(LogSQLChannelAttributes, metaclass=IterableMeta):
	__config_class__ = "Log MSSQL"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class AzureKeyVaultAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Azure Key Vault"
	binding_hostnames = Attribute('Binding Hostnames', min_version='19.2')
	binding_ssl_type = Attribute('Binding SSL Type', min_version='19.2')
	certificate_credential = Attribute('Certificate Credential', min_version='17.2')
	certificate_name = Attribute('Certificate Name', min_version='17.2')
	client_id = Attribute('Client ID', min_version='17.2')
	create_binding = Attribute('Create Binding', min_version='19.2')
	create_san_dns_bindings = Attribute('Create SAN DNS Bindings', min_version='19.2')
	environment = Attribute('Environment', min_version='20.2')
	file_validation_disabled = Attribute('File Validation Disabled', min_version='17.2')
	network_validation_disabled = Attribute('Network Validation Disabled', min_version='17.2')
	non_exportable = Attribute('Non-Exportable', min_version='17.2')
	timeout = Attribute('Timeout', min_version='17.2')
	update_web_app = Attribute('Update Web App', min_version='19.2')
	vault_name = Attribute('Vault Name', min_version='17.2')
	web_app_name = Attribute('Web App Name', min_version='19.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class TrustAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Trust"
	allow_from = Attribute('Allow From', min_version='19.3')
	allowed_algorithm = Attribute('Allowed Algorithm', min_version='19.3')
	allowed_command = Attribute('Allowed Command', min_version='19.3')
	allowed_vendor_types = Attribute('Allowed Vendor Types', min_version='19.3')
	contact = Attribute('Contact', min_version='19.3')
	deny_from = Attribute('Deny From', min_version='19.3')
	key_bit_strength = Attribute('Key Bit Strength', min_version='19.3')
	maximum_authorizations_per_keyset = Attribute('Maximum Authorizations Per Keyset', min_version='20.1')
	minimum_key_bit_strength = Attribute('Minimum Key Bit Strength', min_version='19.3')
	required_options = Attribute('Required Options', min_version='19.3')
	trust_id = Attribute('Trust Id', min_version='19.3')



import requests
import json
from typing import TYPE_CHECKING
from pytpp.attributes._helper import Attribute
if TYPE_CHECKING:
    from packaging.version import Version


class Session:
    """
    This class is responsible for holding the appropriate headers to authenticate each
    request. It also removes all null values from all data sent to TPP.
    """
    def __init__(self, headers: dict, proxies: dict = None, certificate_path: str = None,
                 key_file_path: str = None, verify_ssl: bool = False, tpp_version: 'Version' = None):
        self.requests = requests
        self.requests.packages.urllib3.disable_warnings()
        self.request_kwargs = {
            'headers': headers,
            'verify': verify_ssl
        }
        if proxies:
            self.request_kwargs['proxies'] = proxies
        if bool(certificate_path) != bool(key_file_path):
            raise ValueError('The path to both the certificate file and the key file must be '
                             'given for certificate authentication.')
        if certificate_path and key_file_path:
            self.request_kwargs['cert'] = (certificate_path, key_file_path)
        self.tpp_version = tpp_version

    def update_headers(self, new_headers):
        self.request_kwargs['headers'].update(new_headers)

    def post(self, url: str, data: dict):
        """
        Sends a POST request to the given URL with the given data.

        Args:
            url: URL endpoint
            data: Dictionary of data

        Returns:
            Returns the raw response returned by the server.
        """
        data = self._remove_null_values_from_dictionary(d=data)
        return self.requests.post(url=url, data=json.dumps(data), **self.request_kwargs)

    def get(self, url: str, params: dict = None):
        """
        Sends a GET request to the given URL with the given URL parameters.

        Args:
            url: URL endpoint
            params: Dictionary of parameters to append to the URL

        Returns:
            Returns the raw response returned by the server.
        """
        params = self._remove_null_values_from_dictionary(d=params)
        return self.requests.get(url=url, params=params, **self.request_kwargs)

    def delete(self, url: str, params: dict = None):
        """
        Sends a DELETE request to the given URL.

        Args:
            url: URL endpoint
            params: Dictionary of parameters to append to the URL

        Returns:
            Returns the raw response returned by the server.
        """
        params = self._remove_null_values_from_dictionary(d=params)
        return self.requests.delete(url, params=params, **self.request_kwargs)

    def put(self, url: str, data: dict):
        """
        Sends a PUT request to the given URL with the given data.

        Args:
            url: URL endpoint
            data: Dictionary of data

        Returns:
            Returns the raw response returned by the server.
        """
        data = self._remove_null_values_from_dictionary(d=data)
        return self.requests.put(url=url, data=json.dumps(data), **self.request_kwargs)

    def _remove_null_values_from_dictionary(self, d: dict):
        """
        Removes all values that are of NoneType recursively within a dictionary.
        """
        if not isinstance(d, dict):
            return d

        for key, value in list(d.items()):
            if self.tpp_version:
                if isinstance(key, Attribute) and key.min_version and self.tpp_version < value.min_version:
                    raise AttributeError(f'The attribute "{key}" is not available until TPP version {key.min_version}.')
                if isinstance(value, Attribute) and value.min_version and self.tpp_version < value.min_version:
                    raise AttributeError(f'The attribute "{value}" is not available until TPP version {value.min_version}.')
                if hasattr(value, '__config_class__'):
                    d[key] = value.__config_class__
            if value is None:
                del d[key]
            elif isinstance(value, dict):
                self._remove_null_values_from_dictionary(value)
        return d



from datetime import datetime, timedelta
from typing import List, Union, TYPE_CHECKING
from pytpp.tools.vtypes import Config
from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import InvalidResultCode
if TYPE_CHECKING:
    from pytpp.tools.vtypes import Identity


class _CredentialBase(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)

    def _create(self, name: str, parent_folder: 'Union[Config.Object, str]', friendly_name: str, values: List[dict], expiration: int,
                description: str = None, encryption_key: str = None, shared: bool = False, contacts: 'List[Union[Identity.Identity, str]]' = None,
                get_if_already_exists: bool = True):
        parent_folder_dn = self._get_dn(parent_folder)
        dn = f'{parent_folder_dn}\\{name}'
        expiration = int((datetime.now() + timedelta(expiration * (365 / 12))).timestamp() * 1000)
        result = self._api.websdk.Credentials.Create.post(
            credential_path=dn,
            friendly_name=friendly_name,
            values=values,
            expiration=expiration,
            description=description,
            encryption_key=encryption_key,
            shared=shared,
            contact=[self._get_prefixed_universal(c) for c in contacts] if contacts else None
        ).result

        if result.code != 1:
            if result.code == 401 and get_if_already_exists:
                return self._get_config_object(object_dn=dn)
            raise InvalidResultCode(code=result.code, code_description=result.credential_result)

        return self._get_config_object(object_dn=dn)

    def delete(self, credential: 'Union[Config.Object, str]'):
        """
        Deletes the credential object.

        Args:
            credential: :ref:`config_object` or :ref:`dn` of the credential object.
        """
        credential_dn = self._get_dn(credential)
        result = self._api.websdk.Credentials.Delete.post(credential_path=credential_dn).result
        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.credential_result)

    def get(self, credential_dn: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            credential_dn: :ref:`dn` of the credential object.
            raise_error_if_not_exists: Raise an exception if the credential :ref:`dn` does not exist.

        Returns:
            :ref:`config_object` of the credential object.
        """
        return self._get_config_object(
            object_dn=credential_dn,
            raise_error_if_not_exists=raise_error_if_not_exists
        )


@feature('Amazon Credential')
class AmazonCredential(_CredentialBase):
    def __init__(self, api):
        super().__init__(api)

    def create_adfs(self, name: str, parent_folder: 'Union[Config.Object, str]', adfs_credential: 'Union[Config.Object, str]', adfs_url: str,
                    role: str, expiration: int = 6, description: str = None, encryption_key: str = None, shared: bool = False,
                    contacts: 'List[Union[Identity.Identity, str]]' = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the credential object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            adfs_credential: :ref:`config_object` or :ref:`dn` of the ADFS username credential.
            adfs_url: ADFS URL.
            role: Role.
            expiration: Number months from today at which the credential expires.
            description: Description of the credential object.
            encryption_key: Encryption Key used to protect the credential data.
            shared: If ``True``, the credential can be shared between multiple objects.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` to be set as contacts.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the credential object.
        """
        values = [
            self._name_type_value( name='Source', type='string', value='ADFS'),
            self._name_type_value(name='AdfsUrl', type='string', value=adfs_url),
            self._name_type_value(name='AdfsCredential', type='string', value=self._get_dn(adfs_credential)),
            self._name_type_value(name='Role', type='string', value=role),
        ]
        return self._create(
            name=name,
            parent_folder=parent_folder,
            friendly_name='Amazon',
            values=values,
            expiration=expiration,
            description=description,
            encryption_key=encryption_key,
            shared=shared,
            contacts=contacts,
            get_if_already_exists=get_if_already_exists
        )

    def create_local(self, name: str, parent_folder: 'Union[Config.Object, str]', access_key: str, secret_key: str, role: str = None,
                     external_id: str = None, expiration: int = 6, description: str = None, encryption_key: str = None,
                     shared: bool = False, contacts: 'List[Union[Identity.Identity, str]]' = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the credential object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            access_key: Access Key.
            secret_key: Secret Key.
            role: Role.
            external_id: External ID.
            expiration: Number months from today at which the credential expires.
            description: Description of the credential object.
            encryption_key: Encryption Key used to protect the credential data.
            shared: If ``True``, the credential can be shared between multiple objects.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` to be set as contacts.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the credential object.
        """
        values = [
            self._name_type_value(name='Source', type='string', value='Local'),
            self._name_type_value(name='AccessKey', type='string', value=access_key),
            self._name_type_value(name='ExternalId', type='string', value=external_id),
            self._name_type_value(name='SecretKey', type='string', value=secret_key),
            self._name_type_value(name='Role', type='string', value=role),
        ]
        return self._create(
            name=name,
            parent_folder=parent_folder,
            friendly_name='Amazon',
            values=values,
            expiration=expiration,
            description=description,
            encryption_key=encryption_key,
            shared=shared,
            contacts=contacts,
            get_if_already_exists=get_if_already_exists
        )


@feature('Certificate Credential')
class CertificateCredential(_CredentialBase):
    def __init__(self, api):
        super().__init__(api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', certificate: str, password: str = None, expiration: int = 6,
               description: str = None, encryption_key: str = None, shared: bool = False, contacts: 'List[Union[Identity.Identity, str]]' = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the credential object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            certificate: Base64-encoded PKCS#12 or PFX certificate.
            password: Password.
            expiration: Number months from today at which the credential expires.
            description: Description of the credential object.
            encryption_key: Encryption Key used to protect the credential data.
            shared: If ``True``, the credential can be shared between multiple objects.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` to be set as contacts.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the credential object.
        """
        values = [
            self._name_type_value(name='Certificate', type='byte[]', value=certificate)
        ]
        if password:
            values.append(self._name_type_value('Password', 'string', value=password))

        return self._create(
            name=name,
            parent_folder=parent_folder,
            friendly_name='Certificate',
            values=values,
            expiration=expiration,
            description=description,
            encryption_key=encryption_key,
            shared=shared,
            contacts=contacts,
            get_if_already_exists=get_if_already_exists
        )


@feature('Google Credential')
class GoogleCredential(_CredentialBase):
    def __init__(self, api):
        super().__init__(api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', json_content: 'str', expiration: int = 6,
               description: str = None, encryption_key: str = None, shared: bool = False, contacts: 'List[Union[Identity.Identity, str]]' = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the credential object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            json_content: JSON content as plain text.
            expiration: Number months from today at which the credential expires.
            description: Description of the credential object.
            encryption_key: Encryption Key used to protect the credential data.
            shared: If ``True``, the credential can be shared between multiple objects.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` to be set as contacts.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the credential object.
        """
        import json
        values = [
            self._name_type_value('Json', 'string', value=json.dumps(json_content)),
        ]

        return self._create(
            name=name,
            parent_folder=parent_folder,
            friendly_name='Google',
            values=values,
            expiration=expiration,
            description=description,
            encryption_key=encryption_key,
            shared=shared,
            contacts=contacts,
            get_if_already_exists=get_if_already_exists
        )


@feature('Generic Credential')
class GenericCredential(_CredentialBase):
    def __init__(self, api):
        super().__init__(api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', generic: str, password: str = None, expiration: int = 6,
               description: str = None, encryption_key: str = None, shared: bool = False, contacts: 'List[Union[Identity.Identity, str]]' = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the credential object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            generic: Unclassified binary data.
            password: Password.
            expiration: Number months from today at which the credential expires.
            description: Description of the credential object.
            encryption_key: Encryption Key used to protect the credential data.
            shared: If ``True``, the credential can be shared between multiple objects.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` to be set as contacts.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the credential object.
        """
        values = [
            self._name_type_value('Generic', 'byte[]', value=generic),
            self._name_type_value('Password', 'string', value=password)
        ]

        return self._create(
            name=name,
            parent_folder=parent_folder,
            friendly_name='Generic',
            values=values,
            expiration=expiration,
            description=description,
            encryption_key=encryption_key,
            shared=shared,
            contacts=contacts,
            get_if_already_exists=get_if_already_exists
        )


@feature('Password Credential')
class PasswordCredential(_CredentialBase):
    def __init__(self, api):
        super().__init__(api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', password: str, expiration: int = 6, description: str = None,
               encryption_key: str = None, shared: bool = False, contacts: 'List[Union[Identity.Identity, str]]' = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the credential object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            password: Password.
            expiration: Number months from today at which the credential expires.
            description: Description of the credential object.
            encryption_key: Encryption Key used to protect the credential data.
            shared: If ``True``, the credential can be shared between multiple objects.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` to be set as contacts.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the credential object.
        """
        values = [
            self._name_type_value('Password', 'string', value=password)
        ]

        return self._create(
            name=name,
            parent_folder=parent_folder,
            friendly_name='Password',
            values=values,
            expiration=expiration,
            description=description,
            encryption_key=encryption_key,
            shared=shared,
            contacts=contacts,
            get_if_already_exists=get_if_already_exists
        )


@feature('Private Key Credential')
class PrivateKeyCredential(_CredentialBase):
    def __init__(self, api):
        super().__init__(api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', private_key: str, username: str, expiration: int = 6,
               description: str = None, encryption_key: str = None, shared: bool = False, contacts: 'List[Union[Identity.Identity, str]]' = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the credential object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            private_key: Base64-encoded (PKCS#8) private key.
            username: Username.
            expiration: Number months from today at which the credential expires.
            description: Description of the credential object.
            encryption_key: Encryption Key used to protect the credential data.
            shared: If ``True``, the credential can be shared between multiple objects.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` to be set as contacts.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the credential object.
        """
        values = [
            self._name_type_value('Key', 'byte[]', value=private_key),
            self._name_type_value('Username', 'string', value=username)
        ]

        return self._create(
            name=name,
            parent_folder=parent_folder,
            friendly_name='PrivateKey',
            values=values,
            expiration=expiration,
            description=description,
            encryption_key=encryption_key,
            shared=shared,
            contacts=contacts,
            get_if_already_exists=get_if_already_exists
        )


@feature('Username/Password Credential')
class UsernamePasswordCredential(_CredentialBase):
    def __init__(self, api):
        super().__init__(api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', username: str, password: str, expiration: int = 6,
               description: str = None, encryption_key: str = None, shared: bool = False, contacts: 'List[Union[Identity.Identity, str]]' = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the credential object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            username: Username.
            password: Password.
            expiration: Number months from today at which the credential expires.
            description: Description of the credential object.
            encryption_key: Encryption Key used to protect the credential data.
            shared: If ``True``, the credential can be shared between multiple objects.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` to be set as contacts.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the credential object.
        """
        values = [
            self._name_type_value(name='Username', type='string', value=username),
            self._name_type_value(name='Password', type='string', value=password)
        ]
        return self._create(
            name=name,
            parent_folder=parent_folder,
            friendly_name='UsernamePassword',
            values=values,
            expiration=expiration,
            description=description,
            encryption_key=encryption_key,
            shared=shared,
            contacts=contacts,
            get_if_already_exists=get_if_already_exists
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.cloud_instance_monitor_driver_base import CloudInstanceMonitorDriverBaseAttributes


class AWSEC2InstanceMonitorAttributes(CloudInstanceMonitorDriverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "AWS EC2 Instance Monitor"
	access_key_credential_dn = Attribute('Access Key Credential DN', min_version='17.1')
	aws_credential_dn = Attribute('Aws Credential DN', min_version='18.3')
	secret_key_credential_dn = Attribute('Secret Key Credential DN', min_version='17.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class EntrustNETCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "EntrustNET CA"
	allow_reissue = Attribute('Allow Reissue', min_version='16.4')
	certificate_block = Attribute('Certificate Block')
	certificate_transparency = Attribute('Certificate Transparency', min_version='18.4')
	enhanced_key_usage = Attribute('Enhanced Key Usage', min_version='18.4')
	interval = Attribute('Interval', min_version='15.4')
	organization = Attribute('Organization')
	retrieval_period = Attribute('Retrieval Period', min_version='15.4')
	use_default_organization = Attribute('Use Default Organization', min_version='18.2')
	username_credential = Attribute('Username Credential')
	web_service_url = Attribute('Web Service URL')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class ConnectDirectAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "ConnectDirect"
	certificate_label = Attribute('Certificate Label')
	certificate_only = Attribute('Certificate Only')
	file_validation_disabled = Attribute('File Validation Disabled')
	network_validation_disabled = Attribute('Network Validation Disabled')
	node_name = Attribute('Node Name')
	protocol = Attribute('Protocol')



# noinspection PyUnresolvedReferences
from pytpp._about import (
    __version__, __author__, __author_email__, __project_name__,
    __project_url__,
)
# noinspection PyUnresolvedReferences
from pytpp.api.authenticate import Authenticate
# noinspection PyUnresolvedReferences
from pytpp.properties.oauth import Scope
# noinspection PyUnresolvedReferences
from pytpp.features.definitions.features import Features
# noinspection PyUnresolvedReferences
from pytpp.features.definitions.attributes import Attributes
# noinspection PyUnresolvedReferences
from pytpp.features.definitions.attribute_values import AttributeValues
# noinspection PyUnresolvedReferences
from pytpp.features.definitions.classes import Classes as ClassNames
# noinspection PyUnresolvedReferences
from pytpp.tools.logger import logger
# noinspection PyUnresolvedReferences
from pytpp.tools import vtypes as Types
# Legacy imports
from pytpp.features.definitions.legacy_attribute_names import AttributeNames as __AN
from pytpp.features.definitions.legacy_classes import Classes as __C

def __getattr__(name):
    if name == 'AttributeNames':
        return __AN
    elif name == 'Classes':
        return __C
    raise ImportError(f'{name} cannot be imported because it does not exist.')




from dataclasses import dataclass


@dataclass
class Result:
    code: int
    credential_result: str


@dataclass
class CredentialInfo:
    class_name: str
    full_name: str


@dataclass
class NameTypeValue:
    name: str
    type: str
    value: str



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class PEMAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "PEM"
	certificate_chain_file = Attribute('Certificate Chain File')
	certificate_file = Attribute('Certificate File')
	file_validation_disabled = Attribute('File Validation Disabled')
	network_validation_disabled = Attribute('Network Validation Disabled')
	overwrite_existing_chain = Attribute('Overwrite Existing Chain')
	pbes2_algorithm = Attribute('PBES2 Algorithm')
	private_key_file = Attribute('Private Key File')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.person import PersonAttributes


class UserAttributes(PersonAttributes, metaclass=IterableMeta):
	__config_class__ = "User"
	creation_date = Attribute('Creation Date')
	password = Attribute('Password')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.credential_base import CredentialBaseAttributes


class AutomaticPasswordCredentialAttributes(CredentialBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Automatic Password Credential"
	self_destruct = Attribute('Self Destruct')
	usage_count = Attribute('Usage Count')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.driver_base import DriverBaseAttributes


class CloudInstanceMonitorDriverBaseAttributes(DriverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Cloud Instance Monitor Driver Base"
	certificate_cleanup_options = Attribute('Certificate Cleanup Options', min_version='17.1')
	certificate_relocation_policy_dn = Attribute('Certificate Relocation Policy DN', min_version='17.1')
	cleanup_after = Attribute('Cleanup After', min_version='17.1')
	cloud_region = Attribute('Cloud Region', min_version='17.1')
	last_run = Attribute('Last Run', min_version='17.1')
	policydn = Attribute('PolicyDN', min_version='17.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class F5LTMAdvancedAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "F5 LTM Advanced"
	advanced_settings_bundle_name = Attribute('Advanced Settings Bundle Name')
	advertised_ca = Attribute('Advertised CA')
	archive_location = Attribute('Archive Location')
	authentication_frequency = Attribute('Authentication Frequency')
	build = Attribute('Build')
	bundle_certificate = Attribute('Bundle Certificate')
	bundle_certificate_collection = Attribute('Bundle Certificate Collection')
	crl = Attribute('CRL')
	certificate_chain_name = Attribute('Certificate Chain Name')
	certificate_name = Attribute('Certificate Name')
	chain_traversal_depth = Attribute('Chain Traversal Depth')
	client_authentication_certificate = Attribute('Client Authentication Certificate')
	config_sync = Attribute('Config Sync')
	connection_attempts = Attribute('Connection Attempts', min_version='16.4')
	delete_previous_cert_and_key = Attribute('Delete Previous Cert And Key')
	device_certificate = Attribute('Device Certificate')
	file_validation_disabled = Attribute('File Validation Disabled')
	fips_key = Attribute('Fips Key')
	force_profile_update = Attribute('Force Profile Update', min_version='16.3')
	install_chain_file = Attribute('Install Chain File')
	last_used_host = Attribute('Last Used Host')
	network_validation_disabled = Attribute('Network Validation Disabled')
	overwrite_certificate = Attribute('Overwrite Certificate')
	overwrite_existing_chain = Attribute('Overwrite Existing Chain')
	parent_ssl_profile_name = Attribute('Parent SSL Profile Name')
	partition = Attribute('Partition')
	previous_certificate = Attribute('Previous Certificate')
	previous_key = Attribute('Previous Key')
	provisioning_to = Attribute('Provisioning To')
	sni_default = Attribute('SNI Default', min_version='16.3')
	sni_server_name = Attribute('SNI Server Name', min_version='16.3')
	ssh_port = Attribute('SSH Port')
	ssl_profile_name = Attribute('SSL Profile Name')
	ssl_profile_type = Attribute('SSL Profile Type')
	server_authentication_certificate = Attribute('Server Authentication Certificate')
	server_authentication_name = Attribute('Server Authentication Name')
	system_id = Attribute('System Id')
	trusted_ca = Attribute('Trusted CA')
	use_advanced_settings = Attribute('Use Advanced Settings')
	use_basic_provisioning = Attribute('Use Basic Provisioning')
	use_rest_api = Attribute('Use REST API', min_version='21.3')
	version = Attribute('Version')
	virtual_server_name = Attribute('Virtual Server Name')
	virtual_server_partition = Attribute('Virtual Server Partition')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.agent_certificate_base import AgentCertificateBaseAttributes
from pytpp.attributes.agent_driver_base import AgentDriverBaseAttributes
from pytpp.attributes.schedule_base import ScheduleBaseAttributes


class AgentCertificateDriverAttributes(AgentCertificateBaseAttributes, AgentDriverBaseAttributes, ScheduleBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Agent Certificate Driver"



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.report_base import ReportBaseAttributes


class ReportClientsAttributes(ReportBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Report:Clients"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class GlobalSignMSSLCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "GlobalSign MSSL CA"
	domain_id = Attribute('Domain ID')
	profile_id = Attribute('Profile ID')
	san_type = Attribute('SAN Type')
	web_service_url = Attribute('Web Service URL')



# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import config as Config
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import certificate as Certificate
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import log as Log
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import identity as Identity
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import credential as Credential
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import client as Client
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import permissions as Permissions
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import metadata as CustomFields
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import secret_store as SecretStore
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import ssh as SSH
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import worfklow as Workflow
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import processing_engines as ProcessingEngines
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import stats as Stats
# noinspection PyUnresolvedReferences
from pytpp.properties.response_objects.dataclasses import system_status as SystemStatus
# noinspection PyUnresolvedReferences
from pytpp.features.workflow import RC as ReasonCode



from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import InvalidResultCode
from pytpp.attributes.policy import PolicyAttributes
from concurrent.futures.thread import ThreadPoolExecutor
from typing import List, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from pytpp.tools.vtypes import Config, Identity


@feature('Folder')
class Folder(FeatureBase):
    def __init__(self, api):
        super().__init__(api)

    def apply_workflow(self, folder: 'Union[Config.Object, str]', workflow: 'Union[Config.Object, str]'):
        """
        Applies a workflow to a folder and all of its subordinate objects. However, a subordinate folder
        may block the workflow.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            workflow: :ref:`config_object` or :ref:`dn` of the workflow object.
        """
        folder_dn = self._get_dn(folder)
        workflow_dn = self._get_dn(workflow)
        result = self._api.websdk.Config.AddValue.post(
            object_dn=folder_dn,
            attribute_name=PolicyAttributes.workflow,
            value=workflow_dn
        )

        result.assert_valid_response()

    def block_workflow(self, folder: 'Union[Config.Object, str]', workflow: 'Union[Config.Object, str]'):
        """
        Blocks a workflow on a folder and all of its subordinate objects. This prevents any parent folders from
        enforcing a workflow on this folder and its subordinate objects.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            workflow: :ref:`config_object` or :ref:`dn` of the workflow object.
        """
        folder_dn = self._get_dn(folder)
        workflow_dn = self._get_dn(workflow)
        result = self._api.websdk.Config.AddValue.post(
            object_dn=folder_dn,
            attribute_name=PolicyAttributes.workflow_block,
            value=workflow_dn
        )

        result.assert_valid_response()

    def clear_policy(self, folder: 'Union[Config.Object, str]', class_name: str, attributes: Union[dict, List[str]]):
        """
        If ``attributes`` are not provided, clears the policy attribute name along with all of its values
        on a folder. If ``attributes`` are provided, then only the corresponding policy attribute values
        will be cleared. No error is thrown if the attribute value doesn't exist to begin with. If the
        same attribute name is defined in any ancestor folder, then this folder will inherit that setting.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            class_name: TPP Class Name for the attributes being locked.
            attributes: Two types are supported:

                * ``list`` of attribute names to be cleared entirely.
                * ``dict`` whose keys are attribute names and whose values are the values to be cleared
                  from the attribute. If the attribute is left empty it is cleared.
        """
        folder_dn = self._get_dn(folder)
        if isinstance(attributes, list):
            for attribute_name in attributes:
                result = self._api.websdk.Config.ClearPolicyAttribute.post(
                    object_dn=folder_dn,
                    attribute_name=attribute_name,
                    class_name=str(class_name)
                ).result

                if result.code != 1:
                    raise InvalidResultCode(code=result.code, code_description=result.config_result)

        elif isinstance(attributes, dict):
            for name, values in attributes.items():
                if not isinstance(values, list):
                    values = [values]

                for value in values:
                    result = self._api.websdk.Config.RemovePolicyValue.post(
                        object_dn=folder_dn,
                        class_name=str(class_name),
                        attribute_name=name,
                        value=value
                    ).result

                    if result.code != 1:
                        raise InvalidResultCode(code=result.code, code_description=result.config_result)
        else:
            raise TypeError(f'Expected attributes to be of type List[str] or Dict, but got {type(attributes)} instead.')

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None, log_server: 'Union[Config.Object, str]' = None,
               engines: 'List[Union[Config.Object, str]]' = None, attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the folder.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            description: Description of the policy folder.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` of the contacts.
            log_server: :ref:`config_object` or name of the log server.
            engines: List of :ref:`config_object` or names of the processing engines for this folder.
            attributes: Attributes pertaining to the folder itself and NOT any of the policyable options.
                In order to set engines on this folder, use :meth:`set_engines`. In order to set policyable
                options on the folder, use :meth:`write_policy`.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the folder object.
        """
        folder_attrs = {
            PolicyAttributes.description: description,
            PolicyAttributes.contact: [self._get_prefixed_universal(c) for c in contacts] if contacts else None,
        }
        if attributes:
            folder_attrs.update(attributes)

        folder = self._config_create(
            name=name,
            parent_folder_dn=self._get_dn(parent_folder),
            config_class=PolicyAttributes.__config_class__,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )
        if log_server:
            self._api.websdk.Config.WritePolicy.post(
                object_dn=folder.dn,
                class_name=PolicyAttributes.__config_class__,
                attribute_name=PolicyAttributes.log_view_server,
                values=[self._get_dn(log_server, parent_dn=r'\VED\Logging')]
            )
        if engines:
            self.set_engines(folder=folder, engines=engines, append_engines=True)
        return folder

    def delete(self, folder: 'Union[Config.Object, str]', recursive: bool = True, delete_owners_from_secrets: bool = True, concurrency: int = 1):
        """
        Deletes the folder. The folder is, by default, deleted recursively. All deleted objects will also be removed from their
        secret associations. If the secret association is then orphaned, then it is deleted.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            recursive: If ``True``, delete all objects recursively.
            delete_owners_from_secrets: If ``True``, the owners will be removed from the associated secrets.
            concurrency: If greater than one a thread pool of this size will be used to delete the owner from the secret store association. If ``delete_owners_from_secrets`` is ``False`` then this has no effect.
        """
        folder_dn = self._get_dn(folder)
        response = self._api.websdk.Config.Enumerate.post(object_dn=folder_dn, recursive=recursive)
        result = response.result
        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)
        all_child_objects = response.objects
        if delete_owners_from_secrets:
            # Must delete all of the secrets first.
            with ThreadPoolExecutor(max_workers=concurrency) as pool:
                pool.map(self._secret_store_delete, all_child_objects)
            self._secret_store_delete(object_dn=folder_dn)
        self._config_delete(object_dn=folder_dn, recursive=recursive)

    def delete_engines(self, folder: 'Union[Config.Object, str]'):
        """
        Deletes all processing engines from the folder.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
        """
        folder_guid = self._get_guid(folder)
        return self._api.websdk.ProcessingEngines.Folder.Guid(folder_guid).delete()

    def get(self, folder_dn: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            folder_dn: :ref:`dn` of the folder.
            raise_error_if_not_exists: Raise an exception if the object :ref:`dn` does not exist.

        Returns:
            :ref:`config_object` of the folder object.
        """
        return self._get_config_object(
            object_dn=folder_dn,
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def get_engines(self, folder: 'Union[Config.Object, str]'):
        """
        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.

        Returns:
            List of all :class:`~.dataclasses.processing_engines.Engine` on the folder.
        """
        folder_guid = self._get_guid(folder)
        return self._api.websdk.ProcessingEngines.Folder.Guid(folder_guid).get().engines

    def search(self, object_name_pattern: str = '*', object_types: List[str] = None, recursive: bool = True,
               starting_dn: str = None):
        """
        Searches for an object with the given object name pattern. The pattern is a regular expression. An object type
        can be supplied to specify the TPP object type, such as 'X509 Certificate'. If a starting :ref:`dn` is given without
        an object type, a search will be performed from the starting DN. This can improve the efficiency of this method.
        However, if both a starting DN and object type is provided, due to limitations of the WebSDK API, a search will
        be performed against the object type first, and then filtered by matches to the starting DN.

        Args:
            object_name_pattern: An expression for filtering DN matches.
            object_types: List of TPP Object Types (also called a Config Classes)
            recursive: Search sub-folders when True
            starting_dn: DN of the folder to begin search

        Returns:
            A list of :ref:`config_object` of the objects found.
        """
        if object_types:
            objects = self._api.websdk.Config.FindObjectsOfClass.post(
                classes=object_types,
                pattern=object_name_pattern,
                object_dn=starting_dn,
                recursive=recursive
            ).objects
        elif starting_dn:
            objects = self._api.websdk.Config.Enumerate.post(
                object_dn=starting_dn,
                pattern=object_name_pattern,
                recursive=recursive
            ).objects
        else:
            objects = self._api.websdk.Config.EnumerateAll.post(
                pattern=object_name_pattern
            ).objects

        return objects

    def set_engines(self, folder: 'Union[Config.Object, str]', engines: 'List[Union[Config.Object, str]]',
                    append_engines: bool = False):
        """
        Sets ``engines`` as processing engines for the folder.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            engines: List of engine :ref:`config_object` or engine names listed in TPP.
            append_engines: If ``True``, append ``engines`` to the current list on the folder. Otherwise
                overwrite the current setting.
        """
        folder_guid = self._get_guid(folder)
        engine_guids = [self._get_guid(e, parent_dn=r'\VED\Engines') for e in engines]
        if append_engines:
            current_engines = self._api.websdk.ProcessingEngines.Folder.Guid(folder_guid).get().engines
            engine_guids.extend([engine.engine_guid for engine in current_engines])
        result = self._api.websdk.ProcessingEngines.Folder.Guid(folder_guid).put(engine_guids=engine_guids)
        result.assert_valid_response()

    def read_policy(self, folder: 'Union[Config.Object, str]', class_name: str, attribute_name: str):
        """
        Reads policy settings for the given folder, class name, and attribute name.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            class_name: TPP Class Name for the attributes being locked.
            attribute_name: The attribute name.

        Returns:
            Tuple[List[str], bool]: A tuple of

                * List of values
                * Locked boolean
        """
        folder_dn = self._get_dn(folder)
        resp = self._api.websdk.Config.ReadPolicy.post(
            object_dn=folder_dn,
            class_name=str(class_name),
            attribute_name=attribute_name
        )

        result = resp.result
        if result.code != 1:
            InvalidResultCode(code=result.code, code_description=result.config_result).log()

        return resp.values, resp.locked

    def remove_workflow(self, folder: 'Union[Config.Object, str]', workflow: 'Union[Config.Object, str]'):
        """
        Removes an applied workflow from a folder.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            workflow: :ref:`config_object` or :ref:`dn` of the workflow object.
        """
        folder_dn = self._get_dn(folder)
        workflow_dn = self._get_dn(workflow)
        result = self._api.websdk.Config.RemoveDnValue.post(
            object_dn=folder_dn,
            attribute_name=PolicyAttributes.workflow,
            value=workflow_dn
        )

        result.assert_valid_response()

    def remove_blocked_workflow(self, folder: 'Union[Config.Object, str]', workflow: 'Union[Config.Object, str]'):
        """
        Removes a blocked workflow from a folder.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            workflow: :ref:`config_object` or :ref:`dn` of the workflow object.
        """
        folder_dn = self._get_dn(folder)
        workflow_dn = self._get_dn(workflow)
        result = self._api.websdk.Config.RemoveDnValue.post(
            object_dn=folder_dn,
            attribute_name=PolicyAttributes.workflow_block,
            value=workflow_dn
        )

        result.assert_valid_response()

    def write_policy(self, folder: 'Union[Config.Object, str]', class_name: str, attributes: dict, locked: bool):
        """
        Writes policy settings on a folder. In order to set engines on this folder, use :meth:`set_engines`.
        In order to set custom field policies, use :meth:`pytpp.features.custom_fields.CustomField.write_policy`.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            class_name: TPP Class Name for the attributes being locked.
            attributes: A dictionary of attribute name/value pairs where the name is the
                attribute name and the value is the attribute value.
            locked: Enforces the policy on all subordinate folders and objects.
        """
        folder_dn = self._get_dn(folder)
        for name, values in attributes.items():
            if not isinstance(values, list):
                values = [values]

            result = self._api.websdk.Config.WritePolicy.post(
                object_dn=folder_dn,
                class_name=str(class_name),
                attribute_name=name,
                values=values,
                locked=locked
            ).result

            if result.code != 1:
                InvalidResultCode(code=result.code, code_description=result.config_result).log()

    def update_policy(self, folder: 'Union[Config.Object, str]', class_name: str, attributes: dict, locked: bool):
        """
        Updates policy configurations on a folder.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            class_name: TPP Class Name for the attributes being locked.
            attributes: A dictionary of attribute name/value pairs where the name is the
                attribute name and the value is the attribute value.
            locked: Enforces the policy on all subordinate folders and objects.
        """
        folder_dn = self._get_dn(folder)
        for name, value in attributes.items():
            result = self._api.websdk.Config.AddPolicyValue.post(
                object_dn=folder_dn,
                class_name=str(class_name),
                attribute_name=name,
                value=value,
                locked=locked,
            ).result

            if result.code != 1:
                raise InvalidResultCode(code=result.code, code_description=result.config_result)



from typing import List
from dataclasses import dataclass


@dataclass
class Response:
    success: bool
    error_code: int
    error_message: str


@dataclass
class ProcessingDetails:
    status: str
    status_description: str


@dataclass
class CertificateDetails:
    ca_fingerprint_sha256: str
    certificate_fingerprint_sha256: str
    certificate_type: str
    extensions: dict
    key_id: str
    key_type: str
    principals: List[str]
    public_key_fingerprint_sha256: str
    serial_number: str
    valid_from: int
    valid_to: int


@dataclass
class RequestDetails:
    originating_ip: str
    requested_by: str


@dataclass
class APIClient:
    allowed_to_request_certificate_identifier: bool
    allowed_to_request_extensions: bool
    allowed_to_request_force_command: bool
    allowed_to_request_principals: bool
    allowed_to_request_source_addresses: bool


@dataclass
class Certificate:
    allowed_private_key_algorithms: List[str]
    allowed_private_key_reuse: bool
    certificate_destination_dn: str
    default_private_key_algorithm: str
    signature_hashing_algorithm: str
    type: str
    validity_period: str


@dataclass
class CAKeyPair:
    created_on: str
    dn: str
    fingerprint_sha256: str
    guid: str
    key_algorithm: str
    name: str
    public_key_data: str



from logboss import Logger, LogTagTypes
from logboss.logger import _LoggerPickler

MASKED_REGEXES = [
    '.*password.*',
    '.*token.*',
    '.*private.*key.*',
    'self',
    'cls',
    'apikey',
]

class LogTags:
    api = LogTagTypes.Debug(name='API')
    feature = LogTagTypes.Info(name='Feature')


logger = Logger()
logger.add_log_tags(LogTags)
_LoggerPickler.MASKED_REGEXES = MASKED_REGEXES



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.organization import OrganizationAttributes


class PolicyAttributes(OrganizationAttributes, metaclass=IterableMeta):
	__config_class__ = "Policy"
	certificate_origin = Attribute('Certificate Origin', min_version='19.1')
	log_view_server = Attribute('Log View Server')
	master_preferences = Attribute('Master Preferences', min_version='17.2')
	scep_ca_ident = Attribute('Scep CA Ident')
	scep_certificate_authority = Attribute('Scep Certificate Authority')
	scep_challenge_password = Attribute('Scep Challenge Password')
	scep_encryption_ra_certificate = Attribute('Scep Encryption RA Certificate', min_version='18.3')
	scep_intune_application_id = Attribute('Scep Intune Application Id', min_version='19.3')
	scep_intune_application_secret = Attribute('Scep Intune Application Secret', min_version='19.3')
	scep_intune_tenant_name = Attribute('Scep Intune Tenant Name', min_version='19.3')
	scep_ra_certificate = Attribute('Scep RA Certificate')
	scep_selection_rule = Attribute('Scep Selection Rule')
	scep_signing_ra_certificate = Attribute('Scep Signing RA Certificate', min_version='18.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.report_base import ReportBaseAttributes


class ReportEnhancedCertificateExpirationAttributes(ReportBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Report:EnhancedCertificateExpiration"
	debug_file = Attribute('Debug File')
	grouping = Attribute('Grouping')
	options = Attribute('Options')



# All of these attributes are defined in _about.py, but it must be imported this way.
__project_name__ = None
__version__ = None
__author__ = None
__author_email__= None
__project_url__ = None
exec(open('pytpp/_about.py', 'r').read())

from setuptools import setup, find_packages
import os

PROD_REQUIREMENTS = [
    'isodate==0.6.0',
    'logboss~=0.1.8',
    'packaging==20.9',
    'python-dateutil==2.8.2',
    'requests==2.24.0',
]

DEV_REQUIREMENTS = [
    'pyodbc==4.0.30',
    'lxml==4.4.1',
    'pandas==1.3.3',
    'paramiko==2.7.1'
]

DOC_REQUIREMENTS = [
    'Sphinx>=4.3.2',
    'sphinxcontrib-napoleon~=0.7',
    'sphinx-rtd-theme~=1.0.0',
    'sphinx-rtd-dark-mode~=1.2.4',
]

if __name__ == '__main__':
    with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
        long_description = f.read()

    setup(
        name=__project_name__,
        url=__project_url__,
        version=__version__,
        author=__author__,
        author_email=__author_email__,
        packages=find_packages(include=('pytpp*',)),
        package_dir={
            '': '.'
        },
        description='Venafi TPP Features and WebSDK API In Python',
        long_description=long_description,
        long_description_content_type='text/markdown',
        keywords=['pytpp', 'venafi', 'tpp', 'trust protection platform'],
        install_requires=PROD_REQUIREMENTS,
        extras_require={
            'dev': DEV_REQUIREMENTS,
            'doc': DOC_REQUIREMENTS ,
            'all': DOC_REQUIREMENTS + DEV_REQUIREMENTS
        },
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Programming Language :: Python :: 3.8',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    )



from datetime import datetime
from dataclasses import dataclass
from typing import List


@dataclass
class Counter:
    a_name: str
    b_name: str
    c_name: str
    description: str
    name: str
    stats_type: int


@dataclass
class Result:
    key: 'Key'
    value: 'List[Value]'


@dataclass
class Key:
    m_item1: str
    m_item2: str
    m_item3: str


@dataclass
class Value:
    count: int
    sum_value: int
    tag_a: str
    tag_b: str
    tag_c: str
    time_frame: datetime
    type: int



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class UpgradesRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Upgrades Root"
	grouping_id = Attribute('Grouping Id', min_version='20.1')
	origin_version = Attribute('Origin Version', min_version='20.1')
	start_time = Attribute('Start Time', min_version='20.1')
	stop_time = Attribute('Stop Time', min_version='20.1')
	target_version = Attribute('Target Version', min_version='20.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class WorkflowTicketAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Workflow Ticket"
	approval_explanation = Attribute('Approval Explanation')
	approval_from = Attribute('Approval From')
	approval_reason = Attribute('Approval Reason')
	approver_not_found_timestamp = Attribute('Approver Not Found Timestamp', min_version='16.2')
	creation_date = Attribute('Creation Date')
	last_notification = Attribute('Last Notification')
	last_update = Attribute('Last Update')
	owner_object = Attribute('Owner Object')
	scheduled_start = Attribute('Scheduled Start')
	scheduled_stop = Attribute('Scheduled Stop')
	status = Attribute('Status')
	suspended_attribute = Attribute('Suspended Attribute', min_version='18.2')
	updated_by = Attribute('Updated By')
	user_data = Attribute('User Data')
	workflow = Attribute('Workflow')



from pytpp.properties.resultcodes import ResultCodes
from pytpp.properties.response_objects.dataclasses import config


class Config:
    @staticmethod
    def Result(code: int):
        return config.Result(
            code=code,
            config_result=ResultCodes.Config.get(code, 'Unknown')
        )

    @staticmethod
    def NameValues(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return config.NameValues(
            name=response_object.get('Name'),
            values=response_object.get('Values'),
        )

    @staticmethod
    def Object(response_object: dict):
        """
        See Also:
            See :ref:`config_object` and :ref:`dn` for more information.
        """
        if not isinstance(response_object, dict):
            response_object = {}
        return config.Object(
            absolute_guid=response_object.get('AbsoluteGUID'),
            dn=response_object.get('DN'),
            guid=response_object.get('GUID'),
            config_id=response_object.get('Id'),
            name=response_object.get('Name'),
            parent=response_object.get('Parent'),
            revision=response_object.get('Revision'),
            type_name=response_object.get('TypeName'),
        )

    @staticmethod
    def Policy(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return config.Policy(
            attribute_name=response_object.get('AttributeName'),
            guid=response_object.get('GUID'),
            property=response_object.get('Property'),
            type_name=response_object.get('TypeName'),
            value_list=response_object.get('ValueList'),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class IIS6Attributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "IIS6"
	exportable = Attribute('Exportable')
	file_validation_disabled = Attribute('File Validation Disabled')
	network_validation_disabled = Attribute('Network Validation Disabled')
	site_id = Attribute('Site Id')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.connection_base import ConnectionBaseAttributes
from pytpp.attributes.top import TopAttributes


class JumpServerAttributes(ConnectionBaseAttributes, TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Jump Server"
	consumers = Attribute('Consumers')
	location = Attribute('Location')
	ssh_connection_string = Attribute('SSH Connection String')
	ssh_version = Attribute('SSH Version')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.x509_certificate_base import X509CertificateBaseAttributes
from pytpp.attributes.x509_certificate_validation import X509CertificateValidationAttributes


class X509CertificateAttributes(X509CertificateBaseAttributes, X509CertificateValidationAttributes, metaclass=IterableMeta):
	__config_class__ = "X509 Certificate"
	acme_account_dn = Attribute('ACME Account DN', min_version='17.2')
	application_group_dn = Attribute('Application Group DN', min_version='19.4')
	geotrusttrueflex_ca_emails = Attribute('GeotrustTrueFlex CA:Emails', min_version='15.3')
	geotrusttrueflex_ca_enrollment_mode = Attribute('GeotrustTrueFlex CA:Enrollment Mode', min_version='15.3')
	geotrusttrueflex_ca_first_pickup_request = Attribute('GeotrustTrueFlex CA:First Pickup Request', min_version='15.3')
	geotrusttrueflex_ca_timestamp = Attribute('GeotrustTrueFlex CA:Timestamp', min_version='15.3')
	microsoft_ca_pool_certificate_authority = Attribute('Microsoft CA Pool:Certificate Authority', min_version='19.1')
	portal_download_count = Attribute('Portal Download Count')
	prohibited_san_types = Attribute('Prohibited SAN Types', min_version='18.3')
	ticket_dn = Attribute('Ticket DN')
	work_dn = Attribute('Work DN')
